# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the user authentication service.
"""
import os
import re
import json
import inspect
import secrets
import logging
import flask_security
import os.path as osp
from ..extensions import db
from sqlalchemy import Column, String, JSON
from flask import request, Blueprint, jsonify
from werkzeug.datastructures import MultiDict
from wtforms import StringField, TextAreaField
from wtforms.validators import ValidationError
from flask_principal import Permission, RoleNeed
from flask_security.models import fsqla_v3 as fsqla
from flask_security.forms import ConfirmRegisterForm, Required, Form
from flask_security.utils import base_render_json, suppress_form_csrf
from flask_security import (
    Security as _Security, SQLAlchemyUserDatastore, current_user as cu,
    auth_required
)

bp = Blueprint('schedula_security', __name__)

log = logging.getLogger(__name__)
# Define models
fsqla.FsModels.set_db_info(db)


def is_admin():
    return Permission(RoleNeed('admin')).can()


class Role(db.Model, fsqla.FsRoleMixin):
    def __repr__(self):
        return f'Role({self.id}) {self.name}'


class User(db.Model, fsqla.FsUserMixin):
    firstname = Column(String(255))
    lastname = Column(String(255))
    avatar = Column(JSON())
    custom_data = Column(JSON())
    settings = Column(JSON())

    def name(self):
        return f'{self.firstname} {self.lastname}'

    def __repr__(self):
        return f'User({self.id}) - {self.firstname} {self.lastname} <{self.email}>'

    def get_security_payload(self):
        return {k: v for k, v in {
            'id': self.id,
            'email': self.email,
            'username': self.username,
            'firstname': self.firstname,
            'lastname': self.lastname,
            'avatar': self.avatar,
            'settings': self.settings,
            'custom_data': self.custom_data
        }.items() if v is not None}


class JSONField(StringField):
    def _value(self):
        return json.dumps(super(JSONField, self)._value())


def validate_json(form, field):
    try:
        json.dumps(field.data)
    except ValueError:
        raise ValidationError("Invalid JSON format.")


_re_base64_image_pattern = re.compile(
    r'^data:image\/[a-zA-Z]+;base64,[A-Za-z0-9+/]+={0,2}$'
)


def is_base64_encoded_image(form, field):
    if field.data and not _re_base64_image_pattern.match(field.data):
        raise ValidationError(
            "Invalid avatar format. It must be a Base64 encoded image URL."
        )


# Setup Flask-Security
class EditForm(Form):
    firstname = StringField('firstname', [Required()])
    lastname = StringField('lastname', [Required()])
    avatar = StringField('avatar', [is_base64_encoded_image])
    custom_data = TextAreaField('custom_data', [validate_json])


class ExtendedConfirmRegisterForm(ConfirmRegisterForm, EditForm):
    pass


@bp.route('/edit', methods=['POST'])
@auth_required()
def edit():
    if request.is_json:
        data = MultiDict(request.get_json())
    else:
        data = request.form
    form = EditForm(data, meta=suppress_form_csrf())
    form.user = cu
    if form.validate_on_submit():
        for k, v in form.data.items():
            setattr(cu, k, v)
        db.session.add(cu)
        db.session.commit()
    return base_render_json(form)


@bp.route('/settings', methods=['POST'])
@auth_required()
def settings():
    if request.is_json:
        data = MultiDict(request.get_json())
    else:
        data = request.form
    cu.settings = data
    db.session.add(cu)
    db.session.commit()
    return jsonify({'user': cu.get_security_payload()})


class Security:
    def __init__(self, app, *args, **kwargs):
        if app is not None:
            self.init_app(app, *args, **kwargs)

    def init_app(self, app, *args, **kwargs):
        app.extensions = getattr(app, 'extensions', {})
        defaults = {
            "SECURITY_PASSWORD_SALT": f'{secrets.SystemRandom().getrandbits(128)}',
            "SECURITY_BLUEPRINT_NAME": 'security',
            "SECURITY_URL_PREFIX": '/user',
            "SECURITY_CONFIRMABLE": True,
            "SECURITY_CHANGEABLE": True,
            "SECURITY_AUTO_LOGIN_AFTER_CONFIRM": False,
            "SECURITY_AUTO_LOGIN_AFTER_RESET": True,
            "SECURITY_POST_CONFIRM_VIEW": '/#login',
            "SECURITY_CONFIRM_ERROR_VIEW": '/#login',
            "SECURITY_REGISTERABLE": True,
            "SECURITY_SEND_REGISTER_EMAIL": True,
            "SECURITY_RECOVERABLE": True,
            "SECURITY_RESET_VIEW": '/#reset',
            "SECURITY_RESET_ERROR_VIEW": '/#login',
            "SECURITY_REDIRECT_BEHAVIOR": 'spa',
            "SECURITY_TRACKABLE": True,
            "REMEMBER_COOKIE_SAMESITE": "strict",
            "SESSION_COOKIE_SAMESITE": "strict",
        }
        for k, v in defaults.items():
            app.config[k] = app.config.get(k, os.environ.get(k, v))
            if isinstance(v, bool):
                app.config[k] = str(app.config[k]).lower() == 'true'

        SECURITY_I18N_DIRNAME = [
            "translations",
            os.environ.get('SECURITY_I18N_DIRNAME', 'translations'),
            osp.join(osp.dirname(__file__), 'translations')
        ]
        SECURITY_I18N_DIRNAME.append(osp.join(
            osp.dirname(inspect.getfile(flask_security)), 'translations'
        ))
        app.config['SECURITY_I18N_DIRNAME'] = app.config.get(
            'SECURITY_I18N_DIRNAME', SECURITY_I18N_DIRNAME
        )
        user_datastore = SQLAlchemyUserDatastore(db, User, Role)
        app.security = _Security(
            app, user_datastore,
            confirm_register_form=ExtendedConfirmRegisterForm,
            register_blueprint=True
        )
        app.register_blueprint(bp, url_prefix=app.config["SECURITY_URL_PREFIX"])
        app.extensions['schedula_security'] = self



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the credit application services.
"""
import uuid
import datetime
from .extensions import db
from flask_security.utils import view_commit
from flask_security import current_user as cu
from sqlalchemy import Column, String, DateTime, JSON
from flask import request, jsonify, Blueprint, after_this_request, current_app

bp = Blueprint('schedula_gdpr', __name__)


class Consent(db.Model):
    __tablename__ = 'consents'
    id = Column(String(36), default=lambda: str(uuid.uuid4()), primary_key=True)
    consents = Column(JSON(), nullable=False)
    created_at = Column(
        DateTime(), nullable=False, default=datetime.datetime.utcnow
    )
    updated_at = Column(
        DateTime(), nullable=True, onupdate=datetime.datetime.utcnow
    )
    created_by = db.Column(
        db.Integer, db.ForeignKey('user.id'), nullable=True,
        default=lambda: getattr(cu, 'id', None)
    )
    updated_by = db.Column(
        db.Integer, db.ForeignKey('user.id'), nullable=True,
        onupdate=lambda: getattr(cu, 'id', None)
    )

    def payload(self):
        return {
            'id': self.id,
            'consents': self.consents
        }

    def __repr__(self):
        return f'Consent - {self.id}'


@bp.route('/consent', methods=['POST'])
def consent():
    data = request.json
    consent_id = data.get('id')
    consents = data.get('consents')

    record = consent_id and Consent.query.get(consent_id)
    if record:
        record.consents = consents
    else:
        record = Consent(id=consent_id or None, consents=consents)
    db.session.add(record)
    db.session.flush()
    after_this_request(view_commit)
    return jsonify(record.payload())


@bp.route('/consent/<consent_id>', methods=['GET'])
def check_consent(consent_id):
    consent = Consent.query.get(consent_id)
    return jsonify(consent and consent.payload())


@bp.route('/files/terms-conditions', methods=['GET'])
@bp.route('/files/cookies-policy', methods=['GET'])
def gdpr_files():
    path = request.path.split('/')[-1]
    return current_app.send_static_file(f'gdpr/{path}.pdf')


class GDPR:
    def __init__(self, app, sitemap, *args, **kwargs):
        if app is not None:
            self.init_app(app, sitemap, *args, **kwargs)

    def init_app(self, app, sitemap, *args, **kwargs):
        app.extensions = getattr(app, 'extensions', {})
        app.register_blueprint(bp, url_prefix='/gdpr')
        app.extensions['schedula_gdpr'] = self
        if 'schedula_admin' in app.extensions:
            admin = app.extensions['schedula_admin']
            for v in (Consent,):
                admin.add_model(v, category="GDPR")



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the export page services.
"""
import io
import glob
import gzip
import json
import zipfile
import os.path as osp
from flask import request, send_file, Blueprint, current_app as ca, url_for

bp = Blueprint('schedula_export', __name__)


def custom_url_for(endpoint, **values):
    """Prepend the configured public path for static files."""
    url = url_for(endpoint, **values)
    if url.startswith('/'):
        url = url[1:]
    return url


@bp.route('/<path:form>', methods=['GET', 'POST'])
def export(form):
    data = request.json
    from ... import static_dir
    files = {}
    for folder in (static_dir, ca.static_folder):
        for fp in glob.glob(osp.join(folder, '**', '*'), recursive=True):
            if osp.isdir(fp) or fp.endswith('.map'):
                continue
            arcname = osp.relpath(fp, folder).replace("\\", "/")
            if arcname.startswith('schedula/forms') and not arcname.startswith(
                    f'schedula/forms/{form}-'
            ):
                continue
            files[f'/static/{arcname}'] = fp
    if 'babel' in ca.extensions:
        for folder in ca.extensions['babel'].translation_directories[::-1]:
            for fp in glob.glob(osp.join(folder, '*', "LC_MESSAGES", '*.po')):
                arcname = fp.replace("\\", "/")[:-3].split('/')
                files[f'/locales/{arcname[-3]}/{arcname[-1]}'] = fp

    output_zip = io.BytesIO()
    with zipfile.ZipFile(output_zip, 'w') as zipf:
        zipf.writestr(f'index.html', ca.render_form(form, ctx={
            'url_for': custom_url_for,
            'is_static': 'true'
        }))
        for arcname, fp in files.items():
            if fp.endswith('.gz'):
                with gzip.GzipFile(fp) as gz_file:
                    zipf.writestr(arcname[:-3], gz_file.read())
            else:
                zipf.write(fp, arcname)

        zipf.writestr(
            f'/static/schedula/forms/{form}-data.json', json.dumps(data)
        )
        zipf.writestr(
            f'/locales/languages.json', json.dumps(ca.config.get('BABEL_LANGUAGES'))
        )
        zipf.write(osp.join(osp.dirname(__file__), 'start.py'), '/start.py')
    output_zip.seek(0)

    return send_file(
        output_zip, mimetype='application/zip', as_attachment=True,
        download_name=f'{form}.zip'
    )


class ExportForm:
    def __init__(self, app, sitemap, *args, **kwargs):
        if app is not None:
            self.init_app(app, sitemap, *args, **kwargs)

    def init_app(self, app, sitemap, *args, **kwargs):
        app.render_form = sitemap.render_form
        app.extensions = getattr(app, 'extensions', {})
        app.register_blueprint(bp, url_prefix='/export-form')
        app.extensions['schedula_export_form'] = self



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the base form flask app.

Sub-Modules:

.. currentmodule:: schedula.utils.form.server

.. autosummary::
    :nosignatures:
    :toctree: form/

    admin
    contact
    credits
    csrf
    export
    extensions
    files
    gdpr
    items
    locale
    security
"""
import logging
import schedula as sh

log = logging.getLogger(__name__)


def default_get_form_context():
    from flask import current_app
    from flask_security import current_user as cu
    return {
        'reCAPTCHA': current_app.config.get('RECAPTCHA_PUBLIC_KEY'),
        'stripeKey': current_app.config.get('STRIPE_PUBLISHABLE_KEY'),
        'userInfo': sh.combine_dicts(*(
            getattr(cu, k, lambda: {})()
            for k in ("get_security_payload",)
        )),
    }


def basic_app(sitemap, app):
    from ..config import Config
    from .extensions import db
    app.config.from_object(Config())
    if getattr(sitemap, 'basic_app_config'):
        app.config.from_object(sitemap.basic_app_config)

    # Create database connection object
    db.init_app(app)

    if app.config['SCHEDULA_LOCALE_ENABLED']:
        from .locale import Locales
        Locales(app)

    if app.config['SCHEDULA_CSRF_ENABLED']:
        from .csrf import csrf
        csrf.init_app(app)

    if app.config['SECURITY_ENABLED']:
        from .security import Security
        Security(app)

    if app.config.get('ADMIN_ENABLED'):
        from .admin import Admin
        Admin(app)

    if app.config.get('SCHEDULA_CREDITS_ENABLED'):
        from .credits import Credits
        Credits(app, sitemap)

    if app.config.get('CONTACT_ENABLED'):
        from .contact import Contact
        Contact(app)

    if app.config.get('ITEMS_STORAGE_ENABLED'):
        from .items import Items
        Items(app, sitemap)

    if app.config.get('FILES_STORAGE_ENABLED'):
        from .files import Files
        Files(app, sitemap)

    if app.config['SCHEDULA_GDPR_ENABLED']:
        from .gdpr import GDPR
        GDPR(app, sitemap)

    if app.config['SCHEDULA_EXPORT_FORM_ENABLED']:
        from .export import ExportForm
        ExportForm(app, sitemap)

    @app.context_processor
    def inject_application_root():
        return dict(app_root=app.config.get('APPLICATION_ROOT', ''))

    return app



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

r"""
Define the command line interface for web forms.

.. click:: schedula.utils.form.cli:cli
   :prog: schedula form
   :show-nested:

"""
import os
import click
import logging
import click_log
import os.path as osp

log = logging.getLogger('schedula.utils.form.cli')


class _Logger(logging.Logger):
    def setLevel(self, level):
        super(_Logger, self).setLevel(level)
        frmt = "%(asctime)-15s:%(levelname)5.5s:%(name)s:%(message)s"
        logging.basicConfig(level=level, format=frmt)
        rlog = logging.getLogger()
        # because `basicConfig()` does not reconfig root-logger when re-invoked.
        rlog.level = level
        logging.captureWarnings(True)


logger = _Logger('cli')
click_log.basic_config(logger)


@click.group(
    'form', context_settings=dict(help_option_names=['-h', '--help'])
)
def cli():
    """
    schedula forms command line tool.
    """


@cli.command(
    'sample', short_help='Generates sample folder for a web form.'
)
@click.option(
    '--folder', '-f', default='.', required=False,
    help="Main folder where app and src are contained.",
    type=click.Path(writable=True, file_okay=False)
)
@click_log.simple_verbosity_option(logger)
def sample(folder='.'):
    """
    Writes a sample folder OUTPUT_FOLDER.

    OUTPUT_FOLDER: schedula WebForm folder template. [default: .]
    """
    import glob
    import shutil
    base_dir = osp.dirname(__file__)
    sample_dir = osp.join(base_dir, 'sample')
    root_dir = osp.join(folder, 'root')

    for k in ('static', 'templates', 'translations'):
        shutil.copytree(
            osp.join(base_dir, k), osp.join(root_dir, k), dirs_exist_ok=True
        )

    for fp in glob.glob(osp.join(sample_dir, '*'), include_hidden=True):
        if osp.relpath(fp, sample_dir) in ('package-lock.json', 'node_modules'):
            continue
        it = glob.glob(
            osp.join(fp, '**', '*.*'), include_hidden=True, recursive=True
        ) if osp.isdir(fp) else [fp]
        for f in it:
            dst = osp.join(folder, osp.relpath(f, sample_dir))
            if osp.isfile(f) and not osp.isfile(dst):
                os.makedirs(osp.dirname(dst), exist_ok=True)
                shutil.copy(f, dst)


@cli.command('build', short_help='Build main folder `src` files.')
@click.option(
    '--folder', '-f', default='.', required=False,
    help="Main folder where app and src are contained.",
    type=click.Path(writable=True, file_okay=False)
)
@click_log.simple_verbosity_option(logger)
def build(folder='.'):
    """
    Build main folder `src` files.
    """
    import subprocess
    subprocess.run('npm i; npm run build', cwd=folder, shell=True)


@cli.command(
    'watch',
    short_help='Run server in debug with continuous build of main folder `src` '
               'files.',
    context_settings=dict(
        ignore_unknown_options=True, allow_extra_args=True,
    )
)
@click.option(
    '--folder', '-f', default='.', required=False,
    help="Main folder where app and src are contained.",
    type=click.Path(writable=True, file_okay=False)
)
@click.option(
    '--app', '-a',
    help="The Flask application or factory function to load, in the form "
         "'module:name'.",
    default='app:app',
    required=False
)
@click.option(
    '--only-flask',
    help="The Flask application or factory function to load, in the form "
         "'module:name'.",
    is_flag=True,
    required=False
)
@click_log.simple_verbosity_option(logger)
@click.pass_context
def watch(ctx, folder='.', app='app:app', only_flask=False):
    """
    Run server in debug with continuous build of main folder `src` files.

    MAIN_FOLDER: Folder path. [default: .]
    """
    if only_flask:
        from livereload import Server
        from schedula.utils.form.gapp import get_module
        module, code = app.split(':')
        app = eval(code, get_module(module, (folder,)))
        app.debug = True
        options = {
            ctx.args[i][2:].replace('-', '_'): ctx.args[i + 1]
            for i in range(0, len(ctx.args), 2)
        }
        server = Server(app.wsgi_app)
        server.watch(osp.join(folder, 'root', '**', '*.*'))
        server.serve(**options)
    else:
        import subprocess
        subprocess.run(
            f'(npm i; npm run watch) & '
            f'(python {__file__} watch --app {app} --only-flask '
            f'{" ".join(ctx.args)})',
            cwd=folder, shell=True
        )


@cli.command('run', short_help='Run server.', context_settings=dict(
    ignore_unknown_options=True, allow_extra_args=True,
))
@click.option(
    '--folder', '-f', default='.', required=False,
    help="Main folder where app and src are contained.",
    type=click.Path(writable=True, file_okay=False)
)
@click.option(
    '--app', '-a',
    help="The Flask application or factory function to load, in the form "
         "'module:name'.",
    default='app:app',
    required=False
)
@click_log.simple_verbosity_option(logger)
@click.pass_context
def run(ctx, folder='.', app='app:app'):
    """
    Building of src files within MAIN_FOLDER.

    MAIN_FOLDER: Folder path. [default: .]
    """
    from .gapp import Application, get_module
    module, code = app.split(':')
    options = {
        ctx.args[i][2:].replace('-', '_'): ctx.args[i + 1]
        for i in range(0, len(ctx.args), 2)
    }
    if 'SECRET_KEY' not in os.environ:
        import secrets
        os.environ['SECRET_KEY'] = secrets.token_hex(32)
    if 'SECURITY_PASSWORD_SALT' not in os.environ:
        import secrets
        salt = f'{secrets.SystemRandom().getrandbits(128)}'
        os.environ['SECURITY_PASSWORD_SALT'] = salt

    Application(
        app=eval(code, get_module(module, (folder,))),
        **options
    ).run()


if __name__ == '__main__':
    cli()



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build a flask app from a dispatcher.
"""
import gzip
import base64
import logging
import functools
from ..exc import WebResponse
from ..drw import SiteMap, SiteFolder, FolderNode, SiteNode
from ..base import Base

__author__ = 'Vincenzo Arcidiacono <vinci1it2000@gmail.com>'

log = logging.getLogger(__name__)


class FolderNodeWeb(FolderNode):
    node_data = ()  # ('+set_value',)
    node_function = ('+function',)  # ('+input_domain',)
    edge_data = ()

    def _set_value(self):
        if self.type == 'data' and self.node_id in self.folder.dsp.nodes:
            set_value = self.folder.dsp.set_default_value
            yield '', functools.partial(set_value, self.node_id)

    def _function(self):
        name = 'function'
        if name in self.attr:
            yield '' if self.type == 'function' else name, self.attr[name]


class WebFolder(SiteFolder):
    folder_node = FolderNodeWeb
    ext = ''


class WebNode(SiteNode):
    ext = ''

    @property
    def name(self):
        return self.node_id


class WebMap(SiteMap):
    _view = site_index = lambda *args, **kwargs: None
    site_folder = WebFolder
    site_node = WebNode
    include_folders_as_filenames = False
    methods = ['POST']
    subsite_methods = ['GET', 'POST']
    idle_timeout = 600

    def _repr_svg_(self):
        raise NotImplementedError()

    def basic_app(self, root_path, mute=True, blueprint_name=None, **kwargs):
        app = super(WebMap, self).basic_app(
            root_path, mute=mute, blueprint_name=blueprint_name, **kwargs
        )
        app.before_request(self.before_request)
        app.after_request(self.after_request)
        return app

    def api(self, depth=-1, debug=False):
        from flask import Blueprint
        bp = Blueprint('api', __name__)
        context = self.rules(depth=depth, index=False)
        opt = {'methods': self.methods}
        for i, ((node, extra), path) in enumerate(context.items()):
            view = functools.partial(self._func_handler, node.obj)
            if i:
                bp.add_url_rule('/%s' % path, f'root_{path}', view, **opt)
            else:
                bp.add_url_rule('/', 'root', view, **opt)
                bp.add_url_rule('/%s' % path, 'root', **opt)
        if debug:
            bp.register_blueprint(self.sub_site())
        return bp

    def sub_site(self):
        from flask import Blueprint
        bp = Blueprint('subsite', __name__)
        opt = {'methods': self.subsite_methods}
        bp.add_url_rule('/subsite/<key>/', 'root', self._site_proxy, **opt)
        bp.add_url_rule('/subsite/<key>/<path:path>', 'root', **opt)
        bp.add_url_rule('/subsite/<key>/<string:path>', 'root', **opt)
        return bp

    def app(self, root_path=None, depth=-1, mute=False, blueprint_name=None,
            debug=False, **kwargs):
        kwargs.pop('index', None)
        app = self.basic_app(
            root_path, mute=mute, blueprint_name=blueprint_name, **kwargs
        )
        app.register_blueprint(self.api(depth, debug))
        return app

    def render(self, *args, **kwargs):
        raise NotImplementedError()

    def init_debug_subsite(self, func):
        import random
        from flask import url_for
        from string import ascii_lowercase, digits
        key = ''.join(random.choices(ascii_lowercase + digits, k=12))
        upx = url_for('.subsite.root', key=key)
        site = func.plot(workflow=True, view=False).site(
            index=True, url_prefix=upx, blueprint_name='debug',
            idle_timeout=self.idle_timeout
        ).run()
        self.subsites[key] = (site.url, site.shutdown)
        return upx

    @staticmethod
    def before_request():
        from flask import request
        if request.headers.get('Content-Encoding') == 'gzip':
            request.stream = gzip.GzipFile(fileobj=request.stream)

    @staticmethod
    def after_request(response):
        if response.mimetype == 'text/html':
            return response
        from flask import request, current_app, get_flashed_messages
        messages = get_flashed_messages(with_categories=True)
        if messages:
            headers = {}
            messages = current_app.json.dumps(messages).encode('utf8')
            if 'gzip' in request.headers.get('Accept-Encoding', '').lower():
                messages = base64.b64encode(
                    gzip.compress(messages)
                ).decode("utf-8")
                headers['X-Flash-Messages-length'] = len(messages)
                headers['X-Flash-Messages-Encoding'] = 'gzip'
            headers['X-Flash-Messages'] = messages
            response.headers.update(headers)
        return response

    def _func_handler(self, func):
        from ..dsp import selector
        from flask import request, current_app, Response
        resp = None
        data = {}
        try:
            if not (request.is_json or request.get_data()):
                inp = {}
            else:
                inp = request.get_json(force=True)
            data['input'] = inp
            data['return'] = func(*inp.get('args', ()), **inp.get('kwargs', {}))
            if isinstance(data['return'], Response):
                resp = data['return']
        except WebResponse as ex:
            resp = ex.response
        except Exception as ex:
            data['error'] = str(ex)
        headers = {'Content-Type': 'application/json'}
        if resp is None:
            keys = request.args.get('data', 'return,error').split(',')
            keys = [v.strip(' ') for v in keys]
            content = current_app.json.dumps(selector(
                keys, data, allow_miss=True
            )).encode('utf8')

            if 'gzip' in request.headers.get('Accept-Encoding', '').lower():
                content = gzip.compress(content)
                headers['Content-length'] = len(content)
                headers['Content-Encoding'] = 'gzip'
            resp = current_app.make_response(content)

        if request.headers.get('Debug') == 'true' and isinstance(func, Base):
            headers['Debug-Location'] = self.init_debug_subsite(func)
        resp.headers.update(headers)
        return resp

    def _site_proxy(self, key, path=''):
        import requests
        from flask import request, current_app, abort
        key = key.lower()
        host_url = self.subsites.get(key, (None,))[0]
        if not host_url:
            return abort(404)
        try:
            resp = requests.request(
                method=request.method,
                url=f"{'/'.join((host_url, path or ''))}",
                headers={k: v for k, v in request.headers if k != 'Host'},
                data=request.get_data(),
                cookies=request.cookies,
                allow_redirects=False
            )
        except requests.ConnectionError:
            self.subsites.pop(key, None)
            return abort(503)
        excluded_headers = [
            'content-encoding', 'content-length', 'transfer-encoding',
            'connection'
        ]
        headers = [
            (k, v) for k, v in resp.raw.headers.items()
            if k.lower() not in excluded_headers
        ]
        return current_app.response_class(
            resp.content, resp.status_code, headers
        )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It is a patch to sphinx.ext.autosummary.
"""
import warnings
import logging
import os.path as osp
from pathlib import Path
from typing import Any
from sphinx.util.inspect import getall
from sphinx.util.osutil import ensuredir
from sphinx.ext.autosummary import (
    import_by_name, get_documenter, get_rst_suffix, mock, import_ivar_by_name,
    ImportExceptionGroup
)
from sphinx.ext.autosummary.generate import (
    find_autosummary_in_files, AutosummaryRenderer, ModuleScanner, _get_members,
    _get_module_attrs, _get_modules, _split_full_qualified_name
)
from sphinx.locale import __

logger = logging.getLogger(__name__)
warnings.filterwarnings(
    'ignore', category=DeprecationWarning, module='docutils'
)


def generate_autosummary_content(
        name,
        obj,
        parent,
        template,
        template_name,
        imported_members,
        app,
        recursive,
        context,
        modname=None,
        qualname=None,
):
    doc = get_documenter(app, obj, parent)

    ns = {}
    ns.update(context)

    if doc.objtype == 'module':
        scanner = ModuleScanner(app, obj)
        ns['members'] = scanner.scan(imported_members)

        respect_module_all = not app.config.autosummary_ignore_module_all
        imported_members = imported_members or (
                '__all__' in dir(obj) and respect_module_all
        )
        ns['functions'], ns['all_functions'] = _get_members(
            doc, app, obj, {'function'}, imported=imported_members
        )
        ns['classes'], ns['all_classes'] = _get_members(
            doc, app, obj, {'class'}, imported=imported_members
        )
        ns['exceptions'], ns['all_exceptions'] = _get_members(
            doc, app, obj, {'exception'}, imported=imported_members
        )
        ns['attributes'], ns['all_attributes'] = _get_module_attrs(
            name, ns['members']
        )
        ns['dispatchers'], ns['all_dispatchers'] = _get_members(
            doc, app, obj, {'dispatcher'}, imported=imported_members
        )

        ispackage = hasattr(obj, '__path__')
        if ispackage and recursive:
            # Use members that are not modules as skip list, because it would then mean
            # that module was overwritten in the package namespace
            skip = (
                    ns['all_functions']
                    + ns['all_classes']
                    + ns['all_exceptions']
                    + ns['all_attributes']
                    + ns['all_dispatchers']
            )

            # If respect_module_all and module has a __all__ attribute, first get
            # modules that were explicitly imported. Next, find the rest with the
            # get_modules method, but only put in "public" modules that are in the
            # __all__ list
            #
            # Otherwise, use get_modules method normally
            if respect_module_all and '__all__' in dir(obj):
                imported_modules, all_imported_modules = _get_members(
                    doc, app, obj, {'module'}, imported=True
                )
                skip += all_imported_modules
                public_members = getall(obj)
            else:
                imported_modules, all_imported_modules = [], []
                public_members = None

            modules, all_modules = _get_modules(
                obj, skip=skip, name=name, public_members=public_members
            )
            ns['modules'] = imported_modules + modules
            ns['all_modules'] = all_imported_modules + all_modules
    elif doc.objtype == 'class':
        ns['members'] = dir(obj)
        ns['inherited_members'] = set(dir(obj)) - set(obj.__dict__.keys())
        ns['methods'], ns['all_methods'] = _get_members(
            doc, app, obj, {'method'}, include_public={'__init__'}
        )
        ns['attributes'], ns['all_attributes'] = _get_members(
            doc, app, obj, {'attribute', 'property'}
        )

    if modname is None or qualname is None:
        modname, qualname = _split_full_qualified_name(name)

    if doc.objtype in ('method', 'attribute', 'property'):
        ns['class'] = qualname.rsplit('.', 1)[0]

    if doc.objtype == 'class':
        shortname = qualname
    else:
        shortname = qualname.rsplit('.', 1)[-1]

    ns['fullname'] = name
    ns['module'] = modname
    ns['objname'] = qualname
    ns['name'] = shortname

    ns['objtype'] = doc.objtype
    ns['underline'] = len(name) * '='

    if template_name:
        return template.render(template_name, ns)
    else:
        return template.render(doc.objtype, ns)


def generate_autosummary_docs(
        sources,
        output_dir=None,
        suffix='.rst',
        base_path=None,
        imported_members=False,
        app=None,
        overwrite=True,
        encoding='utf-8',
):
    """Generate autosummary documentation for the given sources.

    :returns: list of generated files (both new and existing ones)
    """
    assert app is not None, 'app is required'

    showed_sources = sorted(sources)
    if len(showed_sources) > 20:
        showed_sources = showed_sources[:10] + ['...'] + showed_sources[-10:]
    logger.info(__('[autosummary] generating autosummary for: %s'),
                ', '.join(showed_sources))

    if output_dir:
        logger.info(__('[autosummary] writing to %s'), output_dir)

    if base_path is not None:
        sources = [osp.join(base_path, filename) for filename in sources]

    template = AutosummaryRenderer(app)

    # read
    items = find_autosummary_in_files(sources)

    # keep track of new files
    new_files = []
    all_files = []

    filename_map = app.config.autosummary_filename_map

    # write
    for entry in sorted(set(items), key=str):
        if entry.path is None:
            # The corresponding autosummary:: directive did not have
            # a :toctree: option
            continue

        path = output_dir or osp.abspath(entry.path)
        ensuredir(path)

        try:
            name, obj, parent, modname = import_by_name(entry.name)
            qualname = name.replace(modname + '.', '')
        except ImportExceptionGroup as exc:
            try:
                # try to import as an instance attribute
                name, obj, parent, modname = import_ivar_by_name(entry.name)
                qualname = name.replace(modname + '.', '')
            except ImportError as exc2:
                if exc2.__cause__:
                    exceptions: list[BaseException] = [*exc.exceptions,
                                                       exc2.__cause__]
                else:
                    exceptions = [*exc.exceptions, exc2]

                errors = list(
                    {f'* {type(e).__name__}: {e}' for e in exceptions}
                )
                logger.warning(
                    __('[autosummary] failed to import %s.\nPossible hints:\n%s'),
                    entry.name,
                    '\n'.join(errors),
                )
                continue

        context: dict[str, Any] = {**app.config.autosummary_context}

        content = generate_autosummary_content(
            name,
            obj,
            parent,
            template,
            entry.template,
            imported_members,
            app,
            entry.recursive,
            context,
            modname,
            qualname,
        )

        file_path = Path(path, filename_map.get(name, name) + suffix)
        all_files.append(file_path)
        if file_path.is_file():
            with file_path.open(encoding=encoding) as f:
                old_content = f.read()

            if content == old_content:
                continue
            if overwrite:  # content has changed
                with file_path.open('w', encoding=encoding) as f:
                    f.write(content)
                new_files.append(file_path)
        else:
            with open(file_path, 'w', encoding=encoding) as f:
                f.write(content)
            new_files.append(file_path)

    # descend recursively to new files
    if new_files:
        all_files.extend(
            generate_autosummary_docs(
                [str(f) for f in new_files],
                output_dir=output_dir,
                suffix=suffix,
                base_path=base_path,
                imported_members=imported_members,
                app=app,
                overwrite=overwrite,
            )
        )

    return all_files


def process_generate_options(app):
    genfiles = app.config.autosummary_generate

    if genfiles is True:
        env = app.builder.env
        genfiles = [str(env.doc2path(x, base=False)) for x in env.found_docs
                    if osp.isfile(env.doc2path(x))]
    elif genfiles is False:
        pass
    else:
        ext = list(app.config.source_suffix)
        genfiles = [
            genfile + (ext[0] if not genfile.endswith(tuple(ext)) else '')
            for genfile in genfiles]

        for entry in genfiles[:]:
            if not osp.isfile(osp.join(app.srcdir, entry)):
                logger.warning(__('autosummary_generate: file not found: %s'),
                               entry)
                genfiles.remove(entry)

    if not genfiles:
        return

    suffix = get_rst_suffix(app)
    if suffix is None:
        logger.warning(__('autosummary generates .rst files internally. '
                          'But your source_suffix does not contain .rst. Skipped.'))
        return

    imported_members = app.config.autosummary_imported_members
    with mock(app.config.autosummary_mock_imports):
        generate_autosummary_docs(genfiles, suffix=suffix, base_path=app.srcdir,
                                  app=app, imported_members=imported_members,
                                  overwrite=app.config.autosummary_generate_overwrite,
                                  encoding=app.config.source_encoding)


def setup(app):
    app.setup_extension('sphinx.ext.autosummary')
    app.add_config_value('autosummary_skip_members', [], 'html')

    # replace callback process_generate_options of 'builder-inited' event.
    import sphinx.ext.autosummary as mdl
    pgo = mdl.process_generate_options
    event = 'builder-inited'
    listeners = app.events.listeners[event]
    from sphinx.events import EventListener
    for i, event in enumerate(listeners):
        if pgo in event:
            listeners[i] = EventListener(*(
                process_generate_options if e is pgo else e for e in event
            ))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides a Blueprint class to construct a Dispatcher and SubDispatch objects.
"""
from .cst import EMPTY
from .base import _Base
from ..dispatcher import Dispatcher

def _init(obj, memo=None):
    return obj.register(memo=memo) if isinstance(obj, Blueprint) else obj


def _safe_call(fn, *args, memo=None, **kwargs):
    return fn(
        *(_init(a, memo) for a in args),
        **{k: _init(v, memo=memo) for k, v in kwargs.items()}
    )


class Blueprint(_Base):
    """Base Blueprint class."""
    cls = Dispatcher

    def __init__(self, *args, **kwargs):
        _Base.__init__(self)
        self.args = args
        self.kwargs = kwargs
        self.deferred = []

    def __getstate__(self):
        d, keys = self.__dict__, ('args', 'kwargs', 'deferred', 'cls')
        return {k: d[k] for k in keys if k in d}

    def _set_cls(self, cls):
        self.cls = cls
        return self

    def register(self, obj=None, memo=None):
        """
        Creates a :class:`Blueprint.cls` and calls each deferred operation.

        :param obj:
            The initialized object with which to call all deferred operations.
        :type obj: object

        :param memo:
            A dictionary to cache registered Blueprints.
        :type memo: dict[Blueprint,T]

        :return:
            The initialized object.
        :rtype: Blueprint.cls | Blueprint

        **--------------------------------------------------------------------**

        Example::

            >>> import schedula as sh
            >>> blue = sh.BlueDispatcher().add_func(len, ['length'])
            >>> blue.register()
            <schedula.dispatcher.Dispatcher object at ...>
        """
        if memo and self in memo:
            obj = memo[self]
            if obj is not None:
                return obj
        if obj is None:
            obj = _safe_call(self.cls, *self.args, memo=memo, **self.kwargs)

        for method, kwargs in self.deferred:
            _safe_call(getattr(obj, method), memo=memo, **kwargs)

        if memo is not None:
            memo[self] = obj

        return obj

    def extend(self, *blues, memo=None):
        """
        Extends deferred operations calling each operation of given Blueprints.

        :param blues:
            Blueprints or Dispatchers to extend deferred operations.
        :type blues: Blueprint | schedula.dispatcher.Dispatcher

        :param memo:
            A dictionary to cache Blueprints.
        :type memo: dict[T,Blueprint]

        :return:
            Self.
        :rtype: Blueprint

        **--------------------------------------------------------------------**

        Example::

            >>> import schedula as sh
            >>> blue = sh.BlueDispatcher()
            >>> blue.extend(
            ...     BlueDispatcher().add_func(len, ['length']),
            ...     BlueDispatcher().add_func(callable, ['is_callable'])
            ... )
            <schedula.utils.blue.BlueDispatcher object at ...>
        """
        memo = {} if memo is None else memo
        for blue in blues:
            if isinstance(blue, Dispatcher):
                blue = blue.blue(memo=memo)
            for method, kwargs in blue.deferred:
                getattr(self, method)(**kwargs)
        return self

    def __call__(self, *args, **kwargs):
        """Calls the registered Blueprint."""
        return self.register(memo={})(*args, **kwargs)


def _parent_blue(func, memo=None, depth=-1):
    from .dsp import add_args, SubDispatch, partial
    memo = {} if memo is None else memo
    if isinstance(func, partial):
        kw = func.keywords
        return func.__class__(
            *(_parent_blue(v, memo, depth) for v in (func.func,) + func.args),
            **{k: _parent_blue(v, memo, depth) for k, v in kw.items()}
        )
    elif isinstance(func, add_args):
        return func.__class__(*(
            _parent_blue(getattr(func, k), memo, depth) for k in func._args
        ))
    elif isinstance(func, (Dispatcher, SubDispatch)):
        return func.blue(memo, depth)
    return func


class BlueDispatcher(Blueprint):
    """
    Blueprint object is a blueprint of how to construct or extend a Dispatcher.

    **------------------------------------------------------------------------**

    **Example**:

    Create a BlueDispatcher::
    
        >>> import schedula as sh
        >>> blue = sh.BlueDispatcher(name='Dispatcher')

    Add data/function/dispatcher nodes to the dispatcher map as usual::
    
        >>> blue.add_data(data_id='a', default_value=3)
        <schedula.utils.blue.BlueDispatcher object at ...>
        >>> @sh.add_function(blue, True, True, outputs=['c'])
        ... def diff_function(a, b=2):
        ...     return b - a
        ...
        >>> blue.add_function(function=max, inputs=['c', 'd'], outputs=['e'])
        <schedula.utils.blue.BlueDispatcher object at ...>
        >>> from math import log
        >>> sub_blue = sh.BlueDispatcher(name='Sub-Dispatcher')
        >>> sub_blue.add_data(data_id='a', default_value=2).add_function(
        ...    function=log, inputs=['a'], outputs=['b']
        ... )
        <schedula.utils.blue.BlueDispatcher object at ...>
        >>> blue.add_dispatcher(sub_blue, ('a',), {'b': 'f'})
        <schedula.utils.blue.BlueDispatcher object at ...>

    You can set the default values as usual::

        >>> blue.set_default_value(data_id='c', value=1, initial_dist=6)
        <schedula.utils.blue.BlueDispatcher object at ...>

    You can also create a `Blueprint` out of `SubDispatchFunction` and add it to
    the `Dispatcher` as follow::
    
        >>> func = sh.SubDispatchFunction(sub_blue, 'func', ['a'], ['b'])
        >>> blue.add_from_lists(fun_list=[
        ...    dict(function=func, inputs=['a'], outputs=['d']),
        ...    dict(function=func, inputs=['c'], outputs=['g']),
        ... ])
        <schedula.utils.blue.BlueDispatcher object at ...>

    Finally you can create the dispatcher object using the method `new`:
    
    .. dispatcher:: dsp
       :opt: graph_attr={'ratio': '1'}
       :code:

        >>> dsp = blue.register(memo={}); dsp
        <schedula.dispatcher.Dispatcher object at ...>

    Or dispatch, calling the Blueprint object:

    .. dispatcher:: sol
       :opt: graph_attr={'ratio': '1'}
       :code:

        >>> sol = blue({'a': 1}); sol
        Solution({'a': 1, 'b': 2, 'c': 1, 'd': 0.0,
                  'f': 0.0, 'e': 1, 'g': 0.0})
    """

    def __init__(self, dmap=None, name='', default_values=None, raises=False,
                 description='', executor=False):
        kwargs = {
            'dmap': dmap, 'name': name, 'default_values': default_values,
            'raises': raises, 'description': description, 'executor': executor
        }
        super(BlueDispatcher, self).__init__(**kwargs)

    def add_data(self, data_id=None, default_value=EMPTY, initial_dist=0.0,
                 wait_inputs=False, wildcard=None, function=None, callback=None,
                 description=None, filters=None, await_result=None, **kwargs):
        """
        Add a single data node to the dispatcher.

        :param data_id:
            Data node id. If None will be assigned automatically ('unknown<%d>')
            not in dmap.
        :type data_id: str, optional

        :param default_value:
            Data node default value. This will be used as input if it is not
            specified as inputs in the ArciDispatch algorithm.
        :type default_value: T, optional

        :param initial_dist:
            Initial distance in the ArciDispatch algorithm when the data node
            default value is used.
        :type initial_dist: float, int, optional

        :param wait_inputs:
            If True ArciDispatch algorithm stops on the node until it gets all
            input estimations.
        :type wait_inputs: bool, optional

        :param wildcard:
            If True, when the data node is used as input and target in the
            ArciDispatch algorithm, the input value will be used as input for
            the connected functions, but not as output. If it is equal to 2, the
            the data node that cannot be calculated are excluded by the wildcard
            condition.
        :type wildcard: bool, int, optional

        :param function:
            Data node estimation function.
            This can be any function that takes only one dictionary
            (key=function node id, value=estimation of data node) as input and
            return one value that is the estimation of the data node.
        :type function: callable, optional

        :param callback:
            Callback function to be called after node estimation.
            This can be any function that takes only one argument that is the
            data node estimation output. It does not return anything.
        :type callback: callable, optional

        :param description:
            Data node's description.
        :type description: str, optional

        :param filters:
            A list of functions that are invoked after the invocation of the
            main function.
        :type filters: list[function], optional

        :param await_result:
            If True the Dispatcher waits data results before assigning them to
            the solution. If a number is defined this is used as `timeout` for
            `Future.result` method [default: False]. Note this is used when
            asynchronous or parallel execution is enable.
        :type await_result: bool|int|float, optional

        :param kwargs:
            Set additional node attributes using key=value.
        :type kwargs: keyword arguments, optional

        :return:
            Self.
        :rtype: BlueDispatcher
        """
        kwargs.update({
            'data_id': data_id, 'filters': filters, 'wait_inputs': wait_inputs,
            'wildcard': wildcard, 'function': function, 'callback': callback,
            'initial_dist': initial_dist, 'default_value': default_value,
            'description': description, 'await_result': await_result
        })
        self.deferred.append(('add_data', kwargs))
        return self

    def add_function(self, function_id=None, function=None, inputs=None,
                     outputs=None, input_domain=None, weight=None,
                     inp_weight=None, out_weight=None, description=None,
                     filters=None, await_domain=None, await_result=None,
                     **kwargs):
        """
        Add a single function node to dispatcher.

        :param function_id:
            Function node id.
            If None will be assigned as <fun.__name__>.
        :type function_id: str, optional

        :param function:
            Data node estimation function.
        :type function: callable, optional

        :param inputs:
            Ordered arguments (i.e., data node ids) needed by the function.
        :type inputs: list, optional

        :param outputs:
            Ordered results (i.e., data node ids) returned by the function.
        :type outputs: list, optional

        :param input_domain:
            A function that checks if input values satisfy the function domain.
            This can be any function that takes the same inputs of the function
            and returns True if input values satisfy the domain, otherwise
            False. In this case the dispatch algorithm doesn't pass on the node.
        :type input_domain: callable, optional

        :param weight:
            Node weight. It is a weight coefficient that is used by the dispatch
            algorithm to estimate the minimum workflow.
        :type weight: float, int, optional

        :param inp_weight:
            Edge weights from data nodes to the function node.
            It is a dictionary (key=data node id) with the weight coefficients
            used by the dispatch algorithm to estimate the minimum workflow.
        :type inp_weight: dict[str, float | int], optional

        :param out_weight:
            Edge weights from the function node to data nodes.
            It is a dictionary (key=data node id) with the weight coefficients
            used by the dispatch algorithm to estimate the minimum workflow.
        :type out_weight: dict[str, float | int], optional

        :param description:
            Function node's description.
        :type description: str, optional

        :param filters:
            A list of functions that are invoked after the invocation of the
            main function.
        :type filters: list[function], optional

        :param await_domain:
            If True the Dispatcher waits all input results before executing the
            `input_domain` function. If a number is defined this is used as
            `timeout` for `Future.result` method [default: True]. Note this is
            used when asynchronous or parallel execution is enable.
        :type await_domain: bool|int|float, optional

        :param await_result:
            If True the Dispatcher waits output results before assigning them to
            the workflow. If a number is defined this is used as `timeout` for
            `Future.result` method [default: False]. Note this is used when
            asynchronous or parallel execution is enable.
        :type await_result: bool|int|float, optional

        :param kwargs:
            Set additional node attributes using key=value.
        :type kwargs: keyword arguments, optional
        """
        kwargs.update({
            'function_id': function_id, 'inputs': inputs, 'function': function,
            'weight': weight, 'input_domain': input_domain, 'filters': filters,
            'await_result': await_result, 'await_domain': await_domain,
            'out_weight': out_weight, 'description': description,
            'outputs': outputs, 'inp_weight': inp_weight
        })
        self.deferred.append(('add_function', kwargs))
        return self

    def add_func(self, function, outputs=None, weight=None, inputs_kwargs=False,
                 inputs_defaults=False, filters=None, input_domain=None,
                 await_domain=None, await_result=None, inp_weight=None,
                 out_weight=None, description=None, inputs=None,
                 function_id=None, **kwargs):
        """
        Add a single function node to dispatcher.

        :param inputs_kwargs:
            Do you want to include kwargs as inputs?
        :type inputs_kwargs: bool

        :param inputs_defaults:
            Do you want to set default values?
        :type inputs_defaults: bool

        :param function_id:
            Function node id.
            If None will be assigned as <fun.__name__>.
        :type function_id: str, optional

        :param function:
            Data node estimation function.
        :type function: callable, optional

        :param inputs:
            Ordered arguments (i.e., data node ids) needed by the function.
            If None it will take parameters names from function signature.
        :type inputs: list, optional

        :param outputs:
            Ordered results (i.e., data node ids) returned by the function.
        :type outputs: list, optional

        :param input_domain:
            A function that checks if input values satisfy the function domain.
            This can be any function that takes the same inputs of the function
            and returns True if input values satisfy the domain, otherwise
            False. In this case the dispatch algorithm doesn't pass on the node.
        :type input_domain: callable, optional

        :param weight:
            Node weight. It is a weight coefficient that is used by the dispatch
            algorithm to estimate the minimum workflow.
        :type weight: float, int, optional

        :param inp_weight:
            Edge weights from data nodes to the function node.
            It is a dictionary (key=data node id) with the weight coefficients
            used by the dispatch algorithm to estimate the minimum workflow.
        :type inp_weight: dict[str, float | int], optional

        :param out_weight:
            Edge weights from the function node to data nodes.
            It is a dictionary (key=data node id) with the weight coefficients
            used by the dispatch algorithm to estimate the minimum workflow.
        :type out_weight: dict[str, float | int], optional

        :param description:
            Function node's description.
        :type description: str, optional

        :param filters:
            A list of functions that are invoked after the invocation of the
            main function.
        :type filters: list[function], optional

        :param await_domain:
            If True the Dispatcher waits all input results before executing the
            `input_domain` function. If a number is defined this is used as
            `timeout` for `Future.result` method [default: True]. Note this is
            used when asynchronous or parallel execution is enable.
        :type await_domain: bool|int|float, optional

        :param await_result:
            If True the Dispatcher waits output results before assigning them to
            the workflow. If a number is defined this is used as `timeout` for
            `Future.result` method [default: False]. Note this is used when
            asynchronous or parallel execution is enable.
        :type await_result: bool|int|float, optional

        :param kwargs:
            Set additional node attributes using key=value.
        :type kwargs: keyword arguments, optional

        :return:
            Self.
        :rtype: BlueDispatcher
        """
        kwargs.update({
            'function_id': function_id, 'inputs': inputs, 'function': function,
            'weight': weight, 'input_domain': input_domain, 'filters': filters,
            'inputs_kwargs': inputs_kwargs, 'inputs_defaults': inputs_defaults,
            'await_result': await_result, 'await_domain': await_domain,
            'out_weight': out_weight, 'description': description,
            'outputs': outputs, 'inp_weight': inp_weight
        })
        self.deferred.append(('add_func', kwargs))
        return self

    def add_dispatcher(self, dsp, inputs=None, outputs=None, dsp_id=None,
                       input_domain=None, weight=None, inp_weight=None,
                       description=None, include_defaults=False,
                       await_domain=None, inputs_prefix='', outputs_prefix='',
                       **kwargs):
        """
        Add a single sub-dispatcher node to dispatcher.

        :param dsp:
            Child dispatcher that is added as sub-dispatcher node to the parent
            dispatcher.
        :type dsp: BlueDispatcher | Dispatcher | dict[str, list]

        :param inputs:
            Inputs mapping. Data node ids from parent dispatcher to child
            sub-dispatcher. If `None` all child dispatcher nodes are used as
            inputs.
        :type inputs: dict[str, str | list[str]] | tuple[str] |
                      (str, ..., dict[str, str | list[str]])

        :param outputs:
            Outputs mapping. Data node ids from child sub-dispatcher to parent
            dispatcher. If `None` all child dispatcher nodes are used as
            outputs.
        :type outputs: dict[str, str | list[str]] | tuple[str] |
                       (str, ..., dict[str, str | list[str]])

        :param dsp_id:
            Sub-dispatcher node id.
            If None will be assigned as <dsp.name>.
        :type dsp_id: str, optional

        :param input_domain:
            A function that checks if input values satisfy the function domain.
            This can be any function that takes the a dictionary with the inputs
            of the sub-dispatcher node and returns True if input values satisfy
            the domain, otherwise False.

            .. note:: This function is invoked every time that a data node reach
               the sub-dispatcher node.
        :type input_domain: (dict) -> bool, optional

        :param weight:
            Node weight. It is a weight coefficient that is used by the dispatch
            algorithm to estimate the minimum workflow.
        :type weight: float, int, optional

        :param inp_weight:
            Edge weights from data nodes to the sub-dispatcher node.
            It is a dictionary (key=data node id) with the weight coefficients
            used by the dispatch algorithm to estimate the minimum workflow.
        :type inp_weight: dict[str, int | float], optional

        :param description:
            Sub-dispatcher node's description.
        :type description: str, optional

        :param include_defaults:
            If True the default values of the sub-dispatcher are added to the
            current dispatcher.
        :type include_defaults: bool, optional

        :param await_domain:
            If True the Dispatcher waits all input results before executing the
            `input_domain` function. If a number is defined this is used as
            `timeout` for `Future.result` method [default: True]. Note this is
            used when asynchronous or parallel execution is enable.
        :type await_domain: bool|int|float, optional

        :param inputs_prefix:
            Add a prefix to parent dispatcher inputs nodes.
        :type inputs_prefix: str

        :param outputs_prefix:
            Add a prefix to parent dispatcher outputs nodes.
        :type outputs_prefix: str

        :param kwargs:
            Set additional node attributes using key=value.
        :type kwargs: keyword arguments, optional

        :return:
            Self.
        :rtype: BlueDispatcher
        """
        kwargs.update({
            'include_defaults': include_defaults, 'await_domain': await_domain,
            'weight': weight, 'input_domain': input_domain, 'dsp_id': dsp_id,
            'description': description, 'outputs': outputs, 'inputs': inputs,
            'inp_weight': inp_weight, 'dsp': dsp,
            'inputs_prefix': inputs_prefix, 'outputs_prefix': outputs_prefix
        })
        self.deferred.append(('add_dispatcher', kwargs))
        return self

    def add_from_lists(self, data_list=None, fun_list=None, dsp_list=None):
        """
        Add multiple function and data nodes to dispatcher.

        :param data_list:
            It is a list of data node kwargs to be loaded.
        :type data_list: list[dict], optional

        :param fun_list:
            It is a list of function node kwargs to be loaded.
        :type fun_list: list[dict], optional

        :param dsp_list:
            It is a list of sub-dispatcher node kwargs to be loaded.
        :type dsp_list: list[dict], optional

        :return:
            Self.
        :rtype: BlueDispatcher
        """
        kwargs = {
            'data_list': data_list, 'fun_list': fun_list, 'dsp_list': dsp_list
        }
        self.deferred.append(('add_from_lists', kwargs))
        return self

    def set_default_value(self, data_id, value=EMPTY, initial_dist=0.0):
        """
        Set the default value of a data node in the dispatcher.

        :param data_id:
            Data node id.
        :type data_id: str

        :param value:
            Data node default value.

            .. note:: If `EMPTY` the previous default value is removed.
        :type value: T, optional

        :param initial_dist:
            Initial distance in the ArciDispatch algorithm when the data node
            default value is used.
        :type initial_dist: float, int, optional

        :return:
            Self.
        :rtype: BlueDispatcher
        """
        kw = {'data_id': data_id, 'value': value, 'initial_dist': initial_dist}
        self.deferred.append(('set_default_value', kw))
        return self



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
Fixes ImportError for MicroPython.
"""
try:
    from threading import Lock
    from weakref import finalize
    from concurrent.futures import Future
    from concurrent.futures._base import Error
except ImportError:  # MicroPython.
    class Future:
        pass


    class Error:
        pass


    class Lock:
        pass


    # noinspection PyUnusedLocal
    def finalize(*args, **kwargs):
        pass



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the CSRF service.
"""
import hmac
import datetime
from flask import current_app, session, g, request
from itsdangerous import URLSafeTimedSerializer
from flask_wtf.csrf import CSRFProtect, CSRFError, generate_csrf


class CSRF(CSRFProtect):
    def setup_form(self, form):
        """
        Receive the form we're attached to and set up fields.

        The default implementation creates a single field of
        type :attr:`field_class` with name taken from the
        ``csrf_field_name`` of the class meta.

        :param form:
            The form instance we're attaching to.
        :return:
            A sequence of `(field_name, unbound_field)` 2-tuples which
            are unbound fields to be added to the form.
        """
        from wtforms.csrf.core import CSRFTokenField
        meta = form.meta
        field_name = meta.csrf_field_name
        unbound_field = CSRFTokenField(label="CSRF Token", csrf_impl=self)
        return [(field_name, unbound_field)]

    def generate_csrf_token(self, csrf_token_field):
        return generate_csrf()

    def validate_csrf_token(self, form, field):
        super().protect()

    def protect(self):
        try:
            super().protect()
        except CSRFError as ex:
            if (current_app.config['CSRF_AUTO_REFRESH_HEADER'] and
                    ex.description == "The CSRF token has expired."):
                field_name = current_app.config['WTF_CSRF_FIELD_NAME']
                secret_key = current_app.config['WTF_CSRF_SECRET_KEY']

                s = URLSafeTimedSerializer(secret_key, salt="wtf-csrf-token")
                token, ts = s.loads(
                    self._get_csrf_token(), return_timestamp=True
                )

                if not hmac.compare_digest(session[field_name], token):
                    self._error_response("The CSRF tokens do not match.")
                time_limit = current_app.config["WTF_CSRF_TIME_LIMIT"] or 0
                if time_limit >= 0:
                    now = datetime.datetime.now(tz=datetime.timezone.utc)
                    if not (0 <= (now - ts).total_seconds() <= time_limit):
                        g.csrf_refresh = True
                g.csrf_valid = True
            else:
                raise ex

    def add_auto_refresh_header(self, resp):
        if g.get('csrf_refresh') or request.endpoint == 'security.logout':
            token = generate_csrf()
            g.csrf_refresh = False
            if token:
                header = current_app.config['CSRF_AUTO_REFRESH_HEADER']
                resp.headers[header] = token
        return resp

    def init_app(self, app):
        app.config.setdefault("WTF_CSRF_HEADERS",
                              ['X-CSRFToken', 'X-CSRF-Token', 'X-XSRF-Token',
                               'X-Csrf-Token'])
        app.config.setdefault("CSRF_AUTO_REFRESH_HEADER", 'N-CSRF-Token')
        app.config.setdefault("WTF_CSRF_SECRET_KEY", app.secret_key)
        app.config.setdefault("WTF_CSRF_TIME_LIMIT", 3600)
        super().init_app(app)
        if app.config['CSRF_AUTO_REFRESH_HEADER']:
            app.after_request(self.add_auto_refresh_header)


csrf = CSRF()



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build a form flask app from a dispatcher.

Sub-Modules:

.. currentmodule:: schedula.utils.form

.. autosummary::
    :nosignatures:
    :toctree: form/

    cli
    config
    gapp
    json_secrets
    server
"""
import functools
import io
import os
import gzip
import glob
import json
import mimetypes
import webbrowser
import os.path as osp
from ..web import WebMap
from . import json_secrets
from jinja2 import TemplateNotFound
from collections import OrderedDict
from werkzeug.exceptions import NotFound

try:
    from smart_open import open as _open
    from smart_open.compression import NO_COMPRESSION

    _open = functools.partial(_open, compression=NO_COMPRESSION)
except ImportError:
    _open = open
__author__ = 'Vincenzo Arcidiacono <vinci1it2000@gmail.com>'

static_dir = osp.join(osp.dirname(__file__), 'static')

STATIC_CONTEXT = {}


def get_static_context():
    if ~all(
            f'main_{k}' in STATIC_CONTEXT and
            osp.exists(STATIC_CONTEXT[f'main_{k}'])
            for k in ('js', 'css')
    ):
        STATIC_CONTEXT.clear()
        static_context = {
            f'main_{k}': osp.relpath(glob.glob(osp.join(
                static_dir, 'schedula', k, f'main.*.{k}.gz'
            ))[0], osp.join(static_dir, 'schedula')).replace('\\', '/')
            for k in ('js', 'css')
        }
        static_context = {
            k: v[:-3] if v.endswith('.gz') else v for k, v in
            static_context.items()
        }
        STATIC_CONTEXT.update(static_context)
    return STATIC_CONTEXT


def get_form_name(name):
    from flask import current_app
    for k in name.split('/')[:1]:
        for sdir in (current_app.static_folder, static_dir):
            sd = osp.join(sdir, 'schedula', 'forms')
            for j in ('schema', 'ui'):
                if osp.exists(osp.join(sd, f'{k}-{j}.json')):
                    return k
    return 'index'


def get_template(form, context, name=None):
    from flask import render_template
    if name is None:
        name = get_form_name(form)
    try:
        template = f'schedula/{form}.html'
        return render_template(template, name=name, form_id=name, **context)
    except TemplateNotFound:
        form = '/'.join(form.split('/')[:-1])
        if form:
            return get_template(form, context, name)
        # noinspection PyUnresolvedReferences
        return get_template('index', context, name=name)


def send_static_file(
        filename, static_folders, is_form=False):
    from flask import current_app, request, send_file
    filename = f'{filename}'.split('/')
    download_name = filename[-1]
    kw = {
        'conditional': True,
        'download_name': download_name,
        'max_age': current_app.get_send_file_max_age(download_name)
    }
    gzipped = 'gzip' in request.headers.get('Accept-Encoding', '').lower()
    if isinstance(static_folders, (list, tuple)):
        static_folders = OrderedDict(((k, True) for k in static_folders))
    for sdir, immutable in static_folders.items():
        sdir = osp.join(sdir, *filename[:-1])
        for ext in ('.gz', '')[::gzipped and 1 or -1]:
            fn = f'{download_name}{ext}'
            fp = osp.join(sdir, fn)
            try:
                with _open(fp, "rb") as f:
                    if is_form:
                        data = json_secrets.dumps(json.load(f)).encode()
                    else:
                        data = f.read()
                if gzipped != bool(ext):
                    func = gzipped and gzip.compress or gzip.decompress
                    f = io.BytesIO(func(data))
                    fn = gzipped and f'{fn}.gz' or download_name
                else:
                    f = io.BytesIO(data)
                try:
                    kw['last_modified'] = os.stat(fp).st_mtime
                except FileNotFoundError:  # Remote file.
                    pass
                mimetype, encoding = mimetypes.guess_type(fn)

                response = send_file(f, **kw)
                response.cache_control.immutable = immutable
                response.cache_control.public = immutable
                response.cache_control.pop('no_cache', None)
                response.cache_control.pop('no-cache', None)
                if immutable:
                    response.cache_control.max_age = 946080000  # 30 years.
                else:
                    response.cache_control.must_revalidate = True
                    response.cache_control.max_age = 604800  # 1 week.
                    response.cache_control.stale_while_revalidate = 120  # 2 min.
                response.content_type = mimetype
                response.content_encoding = encoding
                return response
            except FileNotFoundError:
                continue
    raise NotFound


class FormMap(WebMap):
    def get_form_context(self):
        from .server import default_get_form_context
        context = default_get_form_context().copy()
        if hasattr(self, '_get_form_context'):
            context.update(self._get_form_context())
        return context

    def _get_form_data(self):
        return

    @staticmethod
    def _view(url, *args, **kwargs):
        webbrowser.open(url)

    def __init__(self):
        super(FormMap, self).__init__()
        self.url_prefix = os.environ.get('SCHEDULA_FORM_URL_PREFIX', '')

    def __getattr__(self, item):
        if item.startswith('get_') and hasattr(self, f'_{item}'):
            attr = getattr(self, f'_{item}')
            if isinstance(attr, dict):
                from flask import request
                attr = attr.get(request.path, attr.get(
                    None, getattr(self.__class__, f'_{item}')
                ))
            if hasattr(attr, '__call__'):
                return attr
            return lambda: attr
        return super(FormMap, self).__getattr__(item)

    def render_form(self, form='index', ctx=None):
        from flask import current_app
        from flask_babel import get_locale
        context = {
            'form': self,
            'app': current_app,
            'get_locale': get_locale
        }
        context.update(get_static_context())

        ref_dir = osp.join(current_app.static_folder, 'schedula', 'props')
        for i in ('js', 'css'):
            k = f'props_{i}'
            for j in (form, 'index'):
                if k in context:
                    continue
                for fp in sorted(glob.glob(osp.join(ref_dir, i, f'{j}.*'))):
                    fp = osp.relpath(fp, ref_dir).replace("\\", " / ")
                    if fp.endswith('.gz'):
                        fp = fp[:-3]
                    context[k] = f'props/{fp}'
                    break
        if ctx is not None:
            context.update(ctx)
        return get_template(form, context)

    @staticmethod
    def send_static_file(filename):
        from flask import current_app
        return send_static_file(f'schedula/{filename}', OrderedDict((
            (current_app.static_folder, False), (static_dir, True)
        )), is_form=filename.startswith('forms'))

    def app(self, root_path=None, depth=1, mute=False, blueprint_name=None,
            index=False, debug=False, **kwargs):
        from flask import Blueprint
        app = self.basic_app(
            root_path, mute=mute, blueprint_name=blueprint_name, **kwargs
        )
        bp = Blueprint(
            'schedula', __name__, template_folder='templates'
        )
        bp.add_url_rule('/', 'render_form', self.render_form)
        bp.add_url_rule('/<string:form>', 'render_form')
        bp.add_url_rule('/<string:form>/', 'render_form')
        bp.add_url_rule('/<path:form>', 'render_form')

        bp.add_url_rule(
            '/static/schedula/<path:filename>', 'static', self.send_static_file
        )
        bp.add_url_rule('/static/schedula/<string:filename>', 'static')
        bp.register_blueprint(self.api(depth, debug))
        app.register_blueprint(bp)
        return app

    def basic_app(self, root_path, mute=True, blueprint_name=None, **kwargs):
        app = super(FormMap, self).basic_app(
            root_path, mute=mute, blueprint_name=blueprint_name, **kwargs
        )
        if blueprint_name is None:
            from .server import basic_app
            app = basic_app(self, app)
        return app



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Dispatcher documenter.
"""
import re
import glob
import hashlib
import inspect
import logging
import collections
import os.path as osp
import schedula as sh
from docutils.parsers.rst import directives
from sphinx.ext.autodoc import DataDocumenter, bool_option
from doctest import DocTestParser, DocTestRunner, NORMALIZE_WHITESPACE, ELLIPSIS
from sphinx.ext.autodoc.directive import AutodocDirective

logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------
# Doctest handling
# ------------------------------------------------------------------------------

_re_doctest = re.compile(r'^\s*>>>', re.M)


def contains_doctest(text):
    try:
        # check if it's valid Python as-is
        compile(text, '<string>', 'exec')
        return False
    except SyntaxError:
        pass
    return bool(_re_doctest.search(text))


# ------------------------------------------------------------------------------
# Auto dispatcher content
# ------------------------------------------------------------------------------


def get_grandfather_content(content, level=2, offset=None):
    if offset is None:
        offset = len(content)
    if content.parent and level:
        return get_grandfather_content(
            content.parent, level - 1, offset=offset + content.parent_offset
        )
    return content[:offset]


def _import_docstring_code_content(documenter):
    content = getattr(documenter, 'content', None)
    if content:
        import textwrap

        def get_code(source, c=''):
            s = "\n%s" % c
            return textwrap.dedent(s.join(map(str, source)))

        is_doctest = contains_doctest(get_code(content))
        if is_doctest:
            code = get_code(get_grandfather_content(content))
        else:
            code = get_code(content, '>>> ')
        return code, content


def _import_docstring(documenter):
    code_content = _import_docstring_code_content(documenter)
    if code_content:
        # noinspection PyBroadException
        try:
            code, content = code_content
            parser = DocTestParser()
            runner = DocTestRunner(
                verbose=False, optionflags=NORMALIZE_WHITESPACE | ELLIPSIS
            )

            glob = {}
            if documenter.modname:
                exec('from %s import *\n' % documenter.modname, glob)

            tests = parser.get_doctest(code, glob, '', '', 0)
            runner.run(tests, clear_globs=False)

            documenter.object = tests.globs[documenter.name]
            documenter.code = content
            documenter.is_doctest = True
            return True
        except Exception:
            pass


def _description(lines, dsp, documenter):
    docstring = dsp.__doc__

    if documenter.objpath and documenter.analyzer:
        attr_docs = documenter.analyzer.find_attr_docs()
        key = ('.'.join(documenter.objpath[:-1]), documenter.objpath[-1])
        if key in attr_docs:
            docstring = attr_docs[key]

    if isinstance(docstring, str):
        docstring = docstring.split('\n') + ['']

    lines.extend(docstring)


def _code(lines, documenter):
    if documenter.code:
        if documenter.is_doctest:
            lines += [row.rstrip() for row in documenter.code]
        else:
            lines.extend(['.. code-block:: python', ''])
            lines.extend(['    %s' % r.rstrip() for r in documenter.code])

        lines.append('')


def _plot(lines, dsp, dot_view_opt, documenter, dsp_opt):
    code_content = _import_docstring_code_content(documenter)
    hashkey = str(sorted(dot_view_opt.items())) + '\n'
    if code_content:
        hashkey += code_content[0]
    else:
        modname, objpath = documenter.modname, documenter.objpath
        if modname:
            hashkey += 'import %s\n' % modname
        if objpath:
            hashkey += 'from %s import %s\n' % (modname, '.'.join(objpath))

    fname = 'dispatcher-%s' % hashlib.sha1(hashkey.encode('utf-8')).hexdigest()
    env = documenter.env
    if osp.isabs(env.config.dispatchers_out_dir):
        dspdir = env.config.dispatchers_out_dir
    else:
        dspdir = osp.join(env.srcdir, env.config.dispatchers_out_dir)
    dspdir = osp.join(dspdir, fname)
    index = dot_view_opt.get('index', False)
    viz_js = dot_view_opt.get('viz', False)
    executor = dot_view_opt.get('executor', 'async')
    if osp.isdir(dspdir):
        fpath = glob.glob(osp.join(dspdir, '*.gv'))[0]
    elif hasattr(dsp, 'plot'):
        smap = dsp.plot(**dot_view_opt)
        folder = next(iter(smap))
        fpath = osp.join(dspdir, '%s.gv' % folder.sitemap.foldername)
        dot = folder.dot(smap.rules(index=index, viz_js=viz_js))
        smap.render(
            directory=dspdir, index=index, viz_js=viz_js, executor=executor
        )
        dot.save(fpath, directory='')
    else:
        return

    dsource = osp.dirname(osp.join(env.srcdir, env.docname))
    p = osp.relpath(fpath, dsource).replace('\\', '/')
    dsp_opt = dsp_opt.copy()
    dsp_opt['graphviz_dot'] = dot_view_opt.get('engine', 'dot')
    if index:
        dsp_opt['index'] = ''

    lines.append('.. dsp:: %s' % p)
    lines.extend('   :{}: {}'.format(k, v) for k, v in dsp_opt.items())
    lines.append('')


def _table_heather(lines, title, dsp_name):
    q = 's' if dsp_name and dsp_name[-1] != 's' else ''
    lines.extend(['.. csv-table:: **%s\'%s %s**' % (dsp_name, q, title), ''])


def _data(lines, dsp):
    if isinstance(dsp, sh.SubDispatch):
        dsp = dsp.dsp

    data = sorted(dsp.data_nodes.items())
    if data:
        _table_heather(lines, 'data', dsp.name)
        from schedula.utils.des import get_summary
        from schedula.utils.alg import _search_node_description
        for k, v in data:
            des, link = _search_node_description(dsp, k)

            link = ':obj:`%s <%s>`' % (_node_name(str(k)), link)
            str_format = u'   "%s", "%s"'

            lines.append(str_format % (
                link.replace('"', "'"),
                get_summary(des.split('\n')).replace('"', '""')
            ))

        lines.append('')


dsp_classes = sh.Dispatcher, sh.SubDispatch, sh.Blueprint


def _functions(lines, dsp, node_type='function'):
    if isinstance(dsp, sh.SubDispatch):
        dsp = dsp.dsp

    def check_fun(node_attr):
        if node_attr['type'] not in ('function', 'dispatcher'):
            return False

        if 'function' in node_attr:
            func = sh.parent_func(node_attr['function'])
            c = isinstance(func, dsp_classes)
            return c if node_type == 'dispatcher' else not c
        return node_attr['type'] == node_type

    fun = [v for v in sorted(dsp.nodes.items()) if check_fun(v[1])]

    if fun:
        from schedula.utils.alg import _search_node_description
        _table_heather(lines, '%ss' % node_type, dsp.name)

        for k, v in fun:
            des, full_name = _search_node_description(dsp, k)
            lines.append(u'   ":func:`%s <%s>`", "%s"' % (
                k.replace('"', '""'), full_name.replace('"', '""'),
                des.replace('"', '""')
            ))
        lines.append('')


def _node_name(name):
    return name.replace('<', '\\<').replace('>', '\\>')


PLOT = object()


def _dsp2dot_option(arg):
    """Used to convert the :dmap: option to auto directives."""

    # noinspection PyUnusedLocal
    def map_args(*args, **kwargs):
        from schedula.utils.base import Base
        a = collections.OrderedDict(inspect.signature(
            Base.plot
        ).bind(None, *args, **kwargs).arguments)
        a.popitem(last=False)
        return a

    kw = eval('map_args(%s)' % arg)

    return kw if kw else PLOT


# ------------------------------------------------------------------------------
# Registration hook
# ------------------------------------------------------------------------------


class DispatcherDocumenter(DataDocumenter):
    """
    Specialized Documenter subclass for dispatchers.
    """
    content_indent = ''
    objtype = 'dispatcher'
    directivetype = 'data'
    option_spec = dict(DataDocumenter.option_spec)
    option_spec.update({
        'description': bool_option,
        'opt': _dsp2dot_option,
        'code': bool_option,
        'data': bool_option,
        'func': bool_option,
        'dsp': bool_option,
        'height': directives.length_or_unitless,
        'width': directives.length_or_percentage_or_unitless,
    })
    default_opt = {
        'depth': 0,
        'view': False
    }
    code = None
    is_doctest = False
    blue_cache = {}

    def get_real_modname(self):
        return self.modname

    @classmethod
    def can_document_member(cls, member, *args, **kwargs):
        return isinstance(member, dsp_classes)

    def add_directive_header(self, sig):
        if not self.code:
            if not self.options.annotation:
                self.options.annotation = ' = %s' % self.object.name
            super(DispatcherDocumenter, self).add_directive_header(sig)

    def parse_name(self):
        return super(DispatcherDocumenter, self).parse_name() or True

    def generate(self, more_content=None, **kw):
        # noinspection PyAttributeOutsideInit
        self.content = kw['more_content'] = more_content
        return super(DispatcherDocumenter, self).generate(**kw)

    def import_object(self, *args, **kwargs):
        if _import_docstring(self):
            return True
        self.is_doctest = False
        self.code = None
        res = super(DispatcherDocumenter, self).import_object(*args, **kwargs)
        if res and isinstance(self.object, sh.Blueprint):
            self.object = self.object.register(memo=self.blue_cache)
        return res

    def format_signature(self):
        return ''

    def add_content(self, more_content, no_docstring=False):
        # noinspection PyUnresolvedReferences
        sourcename = self.get_sourcename()
        dsp = self.object
        opt = self.options

        dot_view_opt = self.default_opt.copy()
        if opt.opt and opt.opt is not PLOT:
            dot_view_opt.update(opt.opt)

        lines = []

        if opt.code:
            _code(lines, self)

        if not opt or opt.des:
            _description(lines, dsp, self)

        _plot(lines, dsp, dot_view_opt, self, {
            k: getattr(opt, k) for k in ('height', 'width')
            if getattr(opt, k) is not None
        })

        if not opt or opt.data:
            _data(lines, dsp)

        if not opt or opt.func:
            _functions(lines, dsp)

        if not opt or opt.dsp:
            _functions(lines, dsp, 'dispatcher')

        for line in lines:
            self.add_line(line, sourcename)


class DispatcherDirective(AutodocDirective):
    _default_flags = set(DispatcherDocumenter.option_spec)

    def __init__(self, name, arguments, options, content, lineno,
                 content_offset, *args, **kwargs):
        content._offset = content_offset
        super(DispatcherDirective, self).__init__(
            name, arguments, options, content, lineno, content_offset, *args,
            **kwargs
        )

        if name == 'dispatcher':
            self.name = 'autodispatcher'


def add_autodocumenter(app, cls):
    logger.debug('[app] adding autodocumenter: %r', cls)
    app.registry.add_documenter(cls.objtype, cls)
    app.add_directive('auto' + cls.objtype, DispatcherDirective)


def setup(app):
    """Setup `dispatcher` Sphinx extension module. """
    app.setup_extension('sphinx.ext.autodoc')
    add_autodocumenter(app, DispatcherDocumenter)
    app.add_directive('dispatcher', DispatcherDirective)
    app.add_config_value('dispatchers_out_dir', '_build/_dispatchers', 'html')



import os.path as osp
import schedula as sh

dsp = sh.Dispatcher(name='form')

# Add your model here.

app = dsp.form(
    directory=osp.join(osp.dirname(__file__), 'root'), run=False, view=False
).app()

if __name__ == '__main__':
    app.run(port=50000)



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides the default Flask App config file.
"""
import os
import secrets

try:
    import flask_security
except ImportError:
    flask_security = None


class Config:
    def __init__(self):
        self.DEBUG = False
        self.WTF_CSRF_CHECK_DEFAULT = os.environ.get(
            "WTF_CSRF_CHECK_DEFAULT", 'false'
        ).lower() == 'true'
        self.SCHEDULA_CSRF_ENABLED = os.environ.get(
            "SCHEDULA_CSRF_ENABLED", 'true'
        ).lower() == 'true'
        self.SCHEDULA_GDPR_ENABLED = os.environ.get(
            "SCHEDULA_GDPR_ENABLED", 'true'
        ).lower() == 'true'
        self.SECURITY_ENABLED = os.environ.get(
            "SECURITY_ENABLED", 'true'
        ).lower() == 'true'
        self.ADMIN_ENABLED = os.environ.get(
            "ADMIN_ENABLED", self.SECURITY_ENABLED and 'true' or 'false'
        ).lower() == 'true'
        self.CONTACT_ENABLED = os.environ.get(
            "CONTACT_ENABLED", 'true'
        ).lower() == 'true'
        self.ITEMS_STORAGE_ENABLED = os.environ.get(
            "ITEMS_STORAGE_ENABLED", 'true'
        ).lower() == 'true'
        self.FILES_STORAGE_ENABLED = os.environ.get(
            "FILES_STORAGE_ENABLED", 'true'
        ).lower() == 'true'
        self.SCHEDULA_CREDITS_ENABLED = os.environ.get(
            "SCHEDULA_CREDITS_ENABLED", 'false'
        ).lower() == 'true'
        self.SCHEDULA_LOCALE_ENABLED = os.environ.get(
            "SCHEDULA_LOCALE_ENABLED", 'true'
        ).lower() == 'true'
        self.SCHEDULA_EXPORT_FORM_ENABLED = os.environ.get(
            "SCHEDULA_EXPORT_FORM_ENABLED", 'false'
        ).lower() == 'true'

        # reCAPTCHA configuration
        self.RECAPTCHA_PUBLIC_KEY = os.environ.get(
            'RECAPTCHA_PUBLIC_KEY', '6LcsgJglAAAAAMm7ilxkhBRevaCAuxlpefYZmxHU'
        )
        self.RECAPTCHA_PRIVATE_KEY = os.environ.get(
            'RECAPTCHA_PRIVATE_KEY', '6LcsgJglAAAAAAbR3aHm2qJS_c3XsGqmC9O816eH'
        )

        # Generate a nice key using secrets.token_urlsafe()
        self.SECRET_KEY = os.environ.get("SECRET_KEY", secrets.token_urlsafe())

        self.MAIL_DEFAULT_SENDER = os.environ.get('MAIL_DEFAULT_SENDER')
        self.MAIL_SERVER = os.environ.get('MAIL_HOST')
        self.MAIL_PORT = int(os.environ.get('MAIL_PORT', 465))
        self.MAIL_USE_SSL = os.environ.get(
            'MAIL_USE_SSL', 'true'
        ).lower() == 'true'
        self.MAIL_USERNAME = os.environ.get('MAIL_USER')
        self.MAIL_PASSWORD = os.environ.get('MAIL_PASSWORD')
        self.DB_DIALECT = os.environ.get('DB_DIALECT', 'sqlite')
        self.DB_USER = os.environ.get('DB_USER', '')
        self.DB_PASSWORD = os.environ.get('DB_PASSWORD', '')
        self.DB_HOST = os.environ.get('DB_HOST', '')
        self.DB_PORT = os.environ.get('DB_PORT', '')
        self.DB_DATABASE = os.environ.get('DB_DATABASE', 'database.db')
        self.DB_SERVER = (
                self.DB_USER and f'{self.DB_USER}:{self.DB_PASSWORD}@' or ''
        )
        self.DB_SERVER += (
                self.DB_HOST and f'{self.DB_HOST}:{self.DB_PORT}' or ''
        )
        self.SQLALCHEMY_DATABASE_URI = os.environ.get(
            'SQLALCHEMY_DATABASE_URI',
            f'{self.DB_DIALECT}://{self.DB_SERVER}/{self.DB_DATABASE}'
        )
        # As of Flask-SQLAlchemy 2.4.0 it is easy to pass in options directly to the
        # underlying engine. This option makes sure that DB connections from the
        # pool are still valid. Important for entire application since
        # many DBaaS options automatically close idle connections.
        self.SQLALCHEMY_ENGINE_OPTIONS = {"pool_pre_ping": True}
        self.SQLALCHEMY_TRACK_MODIFICATIONS = os.environ.get(
            'SQLALCHEMY_TRACK_MODIFICATIONS', 'false'
        ).lower() == 'true'



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the contact service.
"""
import os
import rst2txt
import logging
import datetime
import schedula as sh
from flask import render_template
from docutils.core import publish_string
from flask_security import current_user as cu
from flask_mail import Message, Mail as _Mail
from werkzeug.datastructures import MultiDict
from flask_wtf.recaptcha import RecaptchaField
from flask_babel import get_locale
from flask import flash, Blueprint, redirect, request, current_app as ca
from flask_security.utils import (
    base_render_json, suppress_form_csrf, get_post_action_redirect
)
from flask_security.forms import Required, StringField, Form, EmailField
from .locale import lazy_gettext

log = logging.getLogger(__name__)
bp = Blueprint('contact', __name__)


def prepare_message(boby, subject, recipients, reply_to=None, **kwargs):
    body = publish_string(boby, writer=rst2txt.Writer()).decode()
    html = publish_string(boby, writer_name='html').decode()
    return Message(
        body=body, html=html, subject=subject, recipients=recipients,
        reply_to=reply_to, **kwargs
    )


def _render_template(*paths, **data):
    from jinja2.exceptions import TemplateNotFound
    for path in paths:
        try:
            return render_template(path, **data)
        except TemplateNotFound:
            continue
    raise TemplateNotFound


class Mail(_Mail):
    def send_rst(
            self, to, data, rst=None, reply_to=None, body=None, subject=None,
            language=None, **kwargs):
        from tabulate import tabulate
        if language is None:
            locale = get_locale()
            language = f"{locale.language}_{locale.territory}"
        body = body or _render_template(
            f'schedula/email/{rst}/body/{language}.rst',
            f'schedula/email/{rst}/body.rst',
            tabulate=tabulate, **data
        )
        subject = subject or _render_template(
            f'schedula/email/{rst}/subject/{language}.rst',
            f'schedula/email/{rst}/subject.rst',
            tabulate=tabulate, **data
        )
        recipients = list(sh.stlp(os.environ.get('LOCAL', to)))
        message = prepare_message(
            body, subject, recipients, reply_to=reply_to, **kwargs
        )
        return self.send(message)


class ContactForm(Form):
    name = StringField('name', [Required()])
    email = EmailField('email', [Required()])
    subject = StringField('subject', [Required()])
    message = StringField('message', [Required()])
    recaptcha = RecaptchaField('g-recaptcha-response')


def get_post_contact_redirect():
    return get_post_action_redirect("SCHEDULA_POST_CONTACT_VIEW", request.form)


@bp.route('/contact', methods=['POST'])
def contact():
    data = MultiDict(
        request.get_json()
    ) if request.is_json else request.form
    if cu.is_authenticated:
        if 'email' not in data:
            data['email'] = cu.email
        if 'name' not in data:
            data['name'] = f'{cu.firstname} {cu.lastname}'
    form = ContactForm(data, meta=suppress_form_csrf())
    if form.validate_on_submit():
        try:
            ca.extensions['schedula_mail'].send_rst(
                to=[form.data['email'], ca.config.get('MAIL_DEFAULT_SENDER')],
                rst='contact', reply_to=form.data['email'], data={
                    'user': cu, 'data': data,
                    'created': datetime.datetime.now().strftime(
                        "%d/%m/%Y-%H:%M:%S")
                }
            )
            flash(str(lazy_gettext(
                'Your message has been successfully sent!',
                domain='contact'
            )), 'success')
        except Exception as ex:
            flash(str(lazy_gettext(
                "A system error has occurred. Please try again later. "
                "If the problem persists, contact technical support.",
                domain='contact'
            )), 'error')
            return redirect(request.referrer or '/')

    if ca.extensions["security"]._want_json(request):
        return base_render_json(form)
    return redirect(get_post_contact_redirect())


class Contact(Mail):
    def __init__(self, app):
        super().__init__(app)

    def init_app(self, app):
        super().init_app(app)
        app.config['SCHEDULA_POST_CONTACT_VIEW'] = app.config.get(
            'SCHEDULA_POST_CONTACT_VIEW', os.environ.get(
                'SCHEDULA_POST_CONTACT_VIEW', app.config.get(
                    "APPLICATION_ROOT", "/"
                )
            )
        )

        app.register_blueprint(bp, url_prefix='/mail')
        app.extensions['schedula_mail'] = self



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the admin service.
"""
import flask_admin
from .extensions import db
from sqlalchemy import inspect
from flask_security import current_user
from flask_admin import helpers as admin_helpers
from flask import url_for, redirect, abort
from flask_admin.contrib.sqla import ModelView as _ModelView
from flask_admin.form import SecureForm as _SecureForm


class SecureForm(_SecureForm):
    class Meta:
        @property
        def csrf(self):
            from flask import current_app as ca
            return str(ca.config['SCHEDULA_CSRF_ENABLED']).lower() == 'true'

        @property
        def csrf_field_name(self):
            from flask import current_app as ca
            return ca.config['WTF_CSRF_FIELD_NAME']

        def build_csrf(self, form):
            from .csrf import csrf
            return csrf


class ModelView(_ModelView):
    column_display_pk = True
    column_hide_backrefs = False
    form_base_class = SecureForm

    @property
    def column_list(self):
        return [
            c_attr.key for c_attr in inspect(self.model).mapper.column_attrs
        ]

    def is_accessible(self):
        return (
                current_user.is_active and
                current_user.is_authenticated and
                current_user.has_role('admin')
        )

    def _handle_view(self, name, **kwargs):
        """
        Override builtin _handle_view in order to redirect users when a view is not accessible.
        """
        if not self.is_accessible():
            if current_user.is_authenticated:
                # permission denied
                abort(403)
            else:
                # login
                from flask import request
                return redirect(url_for('security.login', next=request.url))


class Admin(flask_admin.Admin):
    def __init__(
            self, app, name='Admin', index_view=None,
            url=None, endpoint=None, **kwargs):
        super().__init__(
            app=None, name=name, index_view=index_view, endpoint=endpoint,
            url=url, **kwargs
        )
        if app is not None:
            self.init_app(
                app, index_view=index_view, endpoint=endpoint, url=url
            )

    def init_app(self, app, **kwargs):
        super().init_app(app, **kwargs)
        app.extensions = getattr(app, 'extensions', {})
        app.extensions['schedula_admin'] = self

        @app.security.context_processor
        def security_context_processor():
            return dict(
                admin_base_template=self.base_template,
                admin_view=self.index_view,
                h=admin_helpers,
                get_url=url_for
            )

        datastore = app.security.datastore
        for k in ('user_model', 'role_model', 'webauthn_model'):
            if getattr(datastore, k, None):
                self.add_model(getattr(datastore, k), category="Security")

    def add_model(self, model, **kwargs):
        self.add_view(ModelView(model, db.session, **kwargs))



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the language service.
"""
import os
import functools
import os.path as osp
from flask_babel import Babel, Domain
from werkzeug.exceptions import NotFound
from werkzeug.datastructures.accept import LanguageAccept
from flask import (
    jsonify, current_app, request, session, Blueprint, send_from_directory
)

bp = Blueprint('locales', __name__)


@functools.lru_cache()
def get_domain(domain):
    return Domain(domain=domain)


def lazy_gettext(*args, domain=None, **kwargs):
    return get_domain(domain).lazy_gettext(*args, **kwargs)


@bp.route('/<language>/<namespace>', methods=['GET'])
def locales(language, namespace):
    for d in current_app.extensions['babel'].translation_directories:
        try:
            return send_from_directory(
                d, f'{osp.join(language, "LC_MESSAGES", namespace)}.po',
                as_attachment=True
            )
        except NotFound:
            pass
    raise NotFound()


@bp.route('/languages.json', methods=['GET'])
def get_locales():
    return jsonify(current_app.config.get('BABEL_LANGUAGES'))


def _set_locale(languages):
    get = current_app.config.get
    session['locale'] = locale = languages.best_match(
        get('BABEL_LANGUAGES'), get('BABEL_DEFAULT_LOCALE')
    )
    return locale


@bp.route('/<lang>', methods=['POST'])
def set_locale(lang):
    languages = LanguageAccept([(lang, 2)] + request.accept_languages)
    return jsonify({"language": _set_locale(languages)})


@bp.route('/<lang>', methods=['GET'])
def get_locale(lang):
    locale = session.get('locale')
    if not locale:
        return set_locale(lang)
    return jsonify({"language": locale})


def _get_locale():
    return session.get('locale') or _set_locale(request.accept_languages)


class Locales:
    def __init__(self, app, *args, **kwargs):
        if app is not None:
            self.init_app(app, *args, **kwargs)

    def init_app(self, app, *args, locale_selector=_get_locale, **kwargs):
        app.extensions = getattr(app, 'extensions', {})
        app.config['BABEL_DEFAULT_LOCALE'] = app.config.get(
            'BABEL_DEFAULT_LOCALE', os.environ.get(
                'BABEL_DEFAULT_LOCALE', 'en_US'
            )
        )
        app.config['BABEL_LANGUAGES'] = app.config.get(
            'BABEL_LANGUAGES', {
                'af_ZA': {"icon": "🇿🇦", "label": "Afrikaans"},
                'ca_ES': {"icon": "🇪🇸", "label": "Català"},
                'da_DK': {"icon": "🇩🇰", "label": "Dansk"},
                'de_DE': {"icon": "🇩🇪", "label": "Deutsch"},
                'en_US': {"icon": "🇺🇸", "label": "English"},
                'es_ES': {"icon": "🇪🇸", "label": "Español"},
                'eu_ES': {"icon": "🇪🇸", "label": "Euskara"},
                'fr_FR': {"icon": "🇫🇷", "label": "Français"},
                'hu_HU': {"icon": "🇭🇺", "label": "Magyar"},
                'hy_AM': {"icon": "🇦🇲", "label": "Հայերեն"},
                'is_IS': {"icon": "🇮🇸", "label": "Íslenska"},
                'it_IT': {"icon": "🇮🇹", "label": "Italiano"},
                'ja_JP': {"icon": "🇯🇵", "label": "日本語"},
                'nl_NL': {"icon": "🇳🇱", "label": "Nederlands"},
                'pl_PL': {"icon": "🇵🇱", "label": "Polski"},
                'pt_BR': {"icon": "🇧🇷", "label": "Português (Brasil)"},
                'pt_PT': {"icon": "🇵🇹", "label": "Português (Portugal)"},
                'ru_RU': {"icon": "🇷🇺", "label": "Русский"},
                'tr_TR': {"icon": "🇹🇷", "label": "Türkçe"},
                'zh_Hans_CN': {"icon": "🇨🇳", "label": "中文（简体）"}
            }
        )
        directories = []
        it = app.config.get(
            'BABEL_TRANSLATION_DIRECTORIES',
            os.environ.get('BABEL_TRANSLATION_DIRECTORIES', 'translations')
        )
        if isinstance(it, str):
            it = [it]
        for k in it:
            directories.extend(k.split(';'))
        if not directories:
            directories.append('translations')
        tdir = osp.join(osp.dirname(__file__), 'translations')
        if tdir not in directories:
            directories.append(tdir)
        app.config['BABEL_TRANSLATION_DIRECTORIES'] = ';'.join(directories)
        Babel(app, *args, locale_selector=locale_selector, **kwargs)
        app.register_blueprint(bp, url_prefix='/locales')
        app.extensions['locates'] = self



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the database.
"""
from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy()



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the item storage service.
"""
import datetime
import schedula as sh
from .extensions import db
from .security import is_admin
from flask import after_this_request, request, jsonify, Blueprint
from sqlalchemy import (
    Column, String, Integer, DateTime, ForeignKey, JSON
)
from flask_security.utils import view_commit
from flask_security import current_user as cu, auth_required

bp = Blueprint('items', __name__)


def default_name(context):
    return f'Item {context.get_current_parameters()["id"]}'



class Item(db.Model):
    __tablename__ = 'item'
    id = Column(Integer, primary_key=True)
    name = Column(String(255), default=default_name)
    category = Column(String(255))
    data = Column('data', JSON)
    user_id = Column(Integer, ForeignKey('user.id'))
    user = db.relationship('User', foreign_keys=[user_id])
    created_at = Column(DateTime(), default=datetime.datetime.utcnow)
    updated_at = Column(DateTime(), onupdate=datetime.datetime.utcnow)

    def payload(self, data=False):
        res = {
            'id': self.id,
            'name': self.name,
            'created_at': self.created_at,
            'updated_at': self.updated_at
        }
        for k in ('created_at', 'updated_at'):
            if res[k]:
                res[k] = res[k].isoformat()

        if data:
            res['data'] = self.data
        return res

    def __repr__(self):
        return f'Item({self.id}) {self.category} - {self.user.name}'


@bp.route('/<category>', methods=['GET', 'POST'])
@bp.route('/<category>/<int:id_item>', methods=[
    'GET', 'PUT', 'PATCH', 'DELETE'
])
@auth_required()
def item(category, id_item=None):
    args = request.args
    method = request.method
    is_get = method == 'GET'
    kw = {'category': category, 'user_id': cu.id}
    if not is_get:
        kw['data'] = request.get_json()
    if 'name' in args:
        kw['name'] = args.get("name", type=str)

    by = {'category': category, 'user_id': cu.id}
    if id_item is not None:
        by['id'] = kw['id'] = id_item
    if is_admin():
        by.pop('user_id')
    if method == 'POST':  # Create.
        item = Item(**kw)
        db.session.add(item)
        db.session.flush()
        payload = item.payload()
    else:  # Read, Delete, Update/Modify, Update/Replace.
        query = Item.query.filter_by(**by)
        if id_item is None:  # GET
            query = query.order_by(Item.id)
            if 'page' in args and 'per_page' in args:
                pag = db.paginate(
                    query,
                    page=args.get("page", type=int),
                    max_per_page=args.get("per_page", type=int),
                    count=True, error_out=False
                )
                items = [item.payload() for item in pag.items]
                payload = {'page': pag.page, 'items': items,
                           'total': pag.total}
            else:
                items = [item.payload() for item in query.all()]
                payload = {'items': items, 'total': len(items)}
        else:
            item = query.first()
            if method == 'DELETE':
                db.session.delete(item)
            elif method in ('PATCH', 'PUT'):
                if method == 'PATCH':
                    kw['data'] = sh.combine_nested_dicts(
                        item.data, kw['data']
                    )
                for k, v in kw.items():
                    setattr(item, k, v)
                db.session.add(item)
                db.session.flush()
            payload = item.payload(data=is_get)
    is_get or after_this_request(view_commit)
    return jsonify(payload)


class Items:
    def __init__(self, app, *args, **kwargs):
        if app is not None:
            self.init_app(app, *args, **kwargs)

    def init_app(self, app, *args, **kwargs):
        app.extensions = getattr(app, 'extensions', {})
        app.register_blueprint(bp, url_prefix='/item')
        app.extensions['item_storage'] = self
        if 'schedula_admin' in app.extensions:
            admin = app.extensions['schedula_admin']
            for v in (Item,):
                admin.add_model(v, category="Items")



import http.server
import socketserver
import webbrowser

# Step 1: Choose a random available port by using port 0
PORT = 0

# Step 2: Create a simple HTTP request handler
Handler = http.server.SimpleHTTPRequestHandler

# Step 3: Open a socket and bind to the port to get the random port number
with socketserver.TCPServer(("", PORT), Handler) as httpd:
    # Get the port that the server is actually using
    PORT = httpd.server_address[1]

    # Step 4: Construct the URL for localhost and the assigned random port
    url = f"http://localhost:{PORT}/"
    print(f"Serving at {url}")

    # Step 5: Automatically open the browser
    webbrowser.open(url)

    # Step 6: Start the server and keep it running
    httpd.serve_forever()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2024, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
import unittest


# noinspection PyTypeChecker
def _make_suite():
    from .utils.test_dsp import (
        TestDispatcherUtils, TestSubDispatcher, TestSubDispatchFunction,
        TestSubDispatchPipe, TestDispatchPipe
    )
    from .utils.test_alg import TestDispatcherGetSubNode
    from .utils.test_gen import TestUtils
    from .test_dispatcher import (
        TestDispatch, TestPerformance, TestBoundaryDispatch, TestNodeOutput,
        TestShrinkDispatcher, TestPipe, TestSubDMap, TestCreateDispatcher
    )

    suite = unittest.TestSuite()
    suite.addTest(TestDispatcherUtils)
    return suite


if __name__ == '__main__':
    import os
    import sys

    os.environ['EXTRAS'] = os.environ.get('EXTRAS', 'micropython')

    runner = unittest.TestRunner()
    result = runner.run(_make_suite())

    # noinspection PyUnresolvedReferences
    sys.exit(result.failuresNum > 0)



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the credit application services.
"""
import os
import re
import copy
import math
import json
import stripe
import datetime
import itertools
import schedula as sh
from .csrf import csrf
from .extensions import db
from .security import User
from .. import json_secrets
from .security import is_admin
from .locale import lazy_gettext
from sqlalchemy.sql import func, column
from flask_security import current_user as cu, auth_required
from flask import jsonify, flash, Blueprint, abort
from sherlock import Lock
from sqlalchemy import Column, String, Integer, DateTime, JSON, or_, event, desc
from dateutil.relativedelta import relativedelta
from dateutil.rrule import (
    rrule, YEARLY, MONTHLY, WEEKLY, DAILY, HOURLY, MINUTELY, SECONDLY
)

FREQUENCIES = {
    'M': MONTHLY, 'W': WEEKLY, 'D': DAILY, 'Y': YEARLY, 'h': HOURLY,
    'm': MINUTELY, 's': SECONDLY
}
_re_freq = re.compile('^(?P<interval>[1-9]\d*)?(?P<freq>[MWDYhms])$')


def date_range(start_time, end_time, freq):
    d = _re_freq.match(freq).groupdict()
    return itertools.pairwise(rrule(
        freq=FREQUENCIES[d['freq']], dtstart=start_time,
        until=end_time, interval=int(d['interval'] or '1')
    ))


bp = Blueprint('schedula_credits', __name__)
users_wallet = db.Table(
    'users_wallet', db.Model.metadata,
    Column('user_id', Integer, db.ForeignKey('user.id'), primary_key=True),
    Column('wallet_id', Integer, db.ForeignKey('wallet.id'), primary_key=True)
)


class Wallet(db.Model):
    __tablename__ = 'wallet'
    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, db.ForeignKey('user.id'), unique=True)
    user = db.relationship('User', foreign_keys=[user_id])
    users = db.relationship('User', secondary=users_wallet)

    def __repr__(self):
        return f'Wallet({self.id}) {self.user.name}'

    def name(self):
        return f"{self.user.firstname or ''} {self.user.lastname or ''}"

    def lock(self):
        return Lock(f'wallet-{self.id}')

    def subscription(self, day=None, session=db.session):
        from flask import current_app as ca
        subscriptions = {}
        api_key = ca.config['STRIPE_SECRET_KEY']
        products = {}

        for subscription in stripe.Subscription.list(
                customer=user2stripe_customer(), status='active',
                expand=['data.items.data.price'], api_key=api_key
        ).auto_paging_iter():
            subs = {}
            for item in subscription.get('items').data:
                product_id = item.price.product
                if product_id in products:
                    features = products[product_id]
                else:
                    product = stripe.Product.retrieve(
                        product_id, api_key=api_key
                    )
                    products[product_id] = features = {
                        product_id: dict(product.metadata),
                    }
                    for v in stripe.Product.list_features(
                            product_id, api_key=api_key
                    ).data:
                        feat = v.entitlement_feature
                        features[feat.lookup_key] = dict(feat.metadata or {})
                subs.update(features)
                subs[item.price.id] = dict(item.price.metadata)
            subscriptions[subscription.id] = {
                k: v for k, v in subs.items() if v
            }
        return subscriptions

    def balance(self, product=None, day=None, session=db.session):
        day = datetime.datetime.today() if day is None else day
        base = session.query(Txn).with_entities(
            func.sum(Txn.credits).label('total_credits'), column('product')
        ).filter_by(
            wallet_id=self.id,
            **({} if product is None else {"product": product})
        ).filter(Txn.valid_from <= day)
        alive_balance = {
            r.product: r.total_credits for r in base.filter(or_(
                Txn.expired_at == None, Txn.expired_at >= day
            )).group_by(Txn.product).all()
        }
        balance = {k: min(v, alive_balance.get(k, 0)) for k, v in (
            (r.product, r.total_credits)
            for r in base.group_by(Txn.product).all()
        )}
        if product is not None:
            balance = balance.get(product, 0)
        return balance

    def use(self, product, credits, session=db.session, created_by=None):
        assert credits >= 0, 'Credits to be consumed have to be positive.'
        with self.lock():
            assert self.balance(
                product, session=session
            ) >= credits, 'Insufficient balance.'
            if created_by is None:
                created_by = cu.id
            t = Txn(
                wallet_id=self.id, type_id=CHARGE, credits=-credits,
                product=product, created_by=created_by
            )
            session.add(t)
            session.commit()
        return t.id

    def charge(self, product, credits, session=db.session):
        assert credits >= 0, 'Credits to be added have to be positive.'
        with self.lock():
            t = Txn(
                wallet_id=self.id, type_id=CHARGE, credits=credits,
                product=product
            )
            session.add(t)
            session.commit()
        return t.id

    def transfer_to(self, product, credits, to_wallet, session=db.session):
        assert credits >= 0, 'Credits to be transfer have to be positive.'

        tran_from = Txn(
            wallet_id=self.id, type_id=TRANSFER, credits=-credits,
            product=product
        )
        tran_to = Txn(
            wallet_id=to_wallet, type_id=TRANSFER, credits=credits,
            product=product
        )
        to_wallet = session.get(Wallet, to_wallet)
        assert to_wallet, 'Destination wallet not found.'
        assert to_wallet, 'Destination wallet not found.'
        with self.lock(), to_wallet.lock():
            assert self.balance(
                product, session=session
            ) >= credits, 'Insufficient balance.'
            session.add_all([tran_from, tran_to])
            session.commit()
        return tran_from.id, tran_to.id


@bp.route('/balance', methods=['GET'])
@bp.route('/balance/<int:wallet_id>', methods=['GET'])
@auth_required()
def get_balance(wallet_id=None):
    from flask import request
    user_id = request.args.get('user_id', cu.id)
    if not is_admin() and cu.id != user_id:
        abort(403)

    get_wallet(user_id)

    query = Wallet.query.filter(or_(
        Wallet.users.any(id=user_id), Wallet.user_id == user_id
    ))
    if wallet_id is not None:
        query = query.filter_by(wallet_id=wallet_id)

    product = request.args.get('product')
    return jsonify({
        wallet.id: {
            'name': wallet.name(),
            'balance': wallet.balance(product),
            'main': wallet.user_id == user_id
        } for wallet in query.all()
    })


@bp.route('/subscription', methods=['GET'])
@bp.route('/subscription/<int:wallet_id>', methods=['GET'])
@auth_required()
def get_subscription(wallet_id=None):
    from flask import request
    user_id = request.args.get('user_id', cu.id)
    kw = {'id': wallet_id, 'user_id': user_id}
    if not is_admin() and cu.id != user_id:
        abort(403)
    if wallet_id is None:
        kw.pop('id', None)
    return jsonify({
        wallet.id: wallet.subscription()
        for wallet in Wallet.query.filter_by(**kw).all()
    })


class TxnType(db.Model):
    __tablename__ = 'transaction_type'
    id = Column(Integer, primary_key=True)
    name = Column(String(255))

    def __repr__(self):
        return f'{self.name}'


class Txn(db.Model):
    __tablename__ = 'wallet_transaction'
    id = Column(Integer, primary_key=True)
    wallet_id = Column(Integer, db.ForeignKey('wallet.id'), nullable=False)
    wallet = db.relationship('Wallet', foreign_keys=[wallet_id])

    type_id = Column(
        Integer, db.ForeignKey('transaction_type.id'), nullable=False
    )
    type = db.relationship('TxnType', foreign_keys=[type_id])

    credits = Column(Integer, default=0)
    product = Column(String(255))
    discount = Column(Integer, default=0)
    subtotal = Column(Integer, default=0)
    tax = Column(Integer, default=0)
    total = Column(Integer, default=0)
    currency = Column(String(64))
    stripe_id = Column(String(255))
    raw_data = Column('raw_data', JSON)
    expired_at = Column(DateTime())
    valid_from = Column(
        DateTime(), nullable=False, default=datetime.datetime.utcnow
    )
    created_at = Column(
        DateTime(), nullable=False, default=datetime.datetime.utcnow
    )
    updated_at = Column(
        DateTime(), nullable=True, onupdate=datetime.datetime.utcnow
    )
    created_by = db.Column(
        db.Integer, db.ForeignKey('user.id'), nullable=True,
        default=lambda: getattr(cu, 'id', None)
    )
    updated_by = db.Column(
        db.Integer, db.ForeignKey('user.id'), nullable=True,
        onupdate=lambda: getattr(cu, 'id', None)
    )

    def __repr__(self):
        return f'Transaction - {self.id}'

    def update_credits(self, credits, session=db.session, force=False):
        assert credits >= 0, 'Credits update have to be positive.'
        assert force or -self.credits >= credits, \
            'Credits update have to be lower than previous.'
        self.credits = -credits
        session.add(self)
        session.flush()


INF_DATE = datetime.datetime(9999, 12, 31, 23, 59)
PURCHASE = 1
REFUND = 2
USAGE = 3
CHARGE = 4
TRANSFER = 5
SUBSCRIPTION = 6


def insert_transaction_type(target, connection, **kw):
    connection.execute(target.insert(), [
        {'id': PURCHASE, 'name': 'Purchase'},
        {'id': REFUND, 'name': 'Refund'},
        {'id': USAGE, 'name': 'Usage'},
        {'id': CHARGE, 'name': 'Charge'},
        {'id': TRANSFER, 'name': 'Transfer'},
        {'id': SUBSCRIPTION, 'name': 'Subscription'}
    ])


event.listen(
    TxnType.__table__, 'after_create', insert_transaction_type
)


def compute_line_items(quantity, tiers, type='graduated'):
    tiers = sorted(tiers, key=lambda x: x.get('last_unit', float('inf')))
    tiers[-1] = {k: v for k, v in tiers[-1].items() if k != 'last_unit'}
    line_items = []
    if type == 'volume':
        tier = next((
            tier for tier in tiers
            if quantity > tier.get('last_unit', float('inf'))
        ))
        per_unit = tier.get('per_unit')
        if per_unit:
            line_items.append(sh.combine_nested_dicts(per_unit, {
                'quantity': quantity, 'metadata': {'credits': quantity}
            }))
        if tier.get('flat_fee'):
            line_items.append(sh.combine_nested_dicts(tier['flat_fee'], {
                'quantity': quantity, 'metadata': {
                    'credits': 0 if per_unit else quantity
                }
            }))
    else:
        prev_unit = 0
        for tier in tiers:
            last_unit = tier.get('last_unit', float('inf'))

            per_unit = tier.get('per_unit')
            credits = (min(last_unit, quantity) - prev_unit)
            if per_unit:
                line_items.append(sh.combine_nested_dicts(per_unit, {
                    'quantity': credits, 'metadata': {'credits': credits}
                }))

            if tier.get('flat_fee'):
                line_items.append(sh.combine_nested_dicts(tier['flat_fee'], {
                    'quantity': 1, 'metadata': {
                        'credits': 0 if per_unit else credits
                    }
                }))
            if quantity <= last_unit:
                break
            prev_unit = tier['last_unit']
    return line_items


def user2stripe_customer(user=cu, limit=1, **kwargs):
    from flask import current_app as ca
    api_key = ca.config['STRIPE_SECRET_KEY']
    with Lock(f'Stripe-customer-{user.id}'):
        for customer in stripe.Customer.search(
                query=f"email:'{user.email}'", api_key=api_key, limit=1,
                **kwargs
        ).data:
            if customer.metadata.user_id != str(user.id):
                customer = stripe.Customer.modify(
                    customer.id, api_key=api_key,
                    metadata={"user_id": str(user.id)}
                )
            break
        else:
            customer = stripe.Customer.create(
                api_key=api_key, email=user.email,
                name=f'{user.firstname} {user.lastname}',
                metadata={'user_id': str(user.id)}
            )

    return customer


def stripe_customer2user(customer):
    user = User.query.get(customer.metadata['user_id'])
    if user:
        return user
    user = User.query.filter_by(email=customer.email).first()
    if user:
        return user
    from flask import current_app as ca
    user = ca.security.datastore.create_user(
        email=customer.email, firstname=customer.name
    )
    db.session.flush([user])

    api_key = ca.config['STRIPE_SECRET_KEY']
    stripe.Customer.modify(
        customer.id, api_key=api_key, metadata={"user_id": str(user.id)}
    )
    return user


def get_wallet(user_id, session=db.session):
    with Lock(f'wallet-user-{user_id}'):
        wallet = session.query(Wallet).filter_by(user_id=user_id).one_or_none()
        if not wallet:
            wallet = Wallet(user_id=user_id)
            session.add(wallet)
            session.commit()
    return wallet


@bp.route('/create-customer-pricing-table-session', methods=['POST'])
@auth_required()
def create_pricing_table():
    from flask import request, current_app as ca
    try:
        data = request.get_json() if request.is_json else dict(request.form)
        data = json_secrets.secrets(data, False)
        customer = user2stripe_customer()
        session = stripe.CustomerSession.create(
            api_key=ca.config['STRIPE_SECRET_KEY'],
            **sh.combine_nested_dicts(data, base={
                "customer": customer.id,
                "components": {"pricing_table": {"enabled": True}}
            })
        )
    except Exception as e:
        return jsonify(error=str(e))

    return jsonify(clientSecret=session.client_secret)


@bp.route('/create-customer-portal-session', methods=['POST'])
@auth_required()
def create_portal(skip_data=False):
    from stripe.billing_portal import Session
    from flask import request, current_app as ca, session
    try:
        if skip_data:
            data = {}
        else:
            data = request.get_json() if request.is_json else dict(request.form)
            data = json_secrets.secrets(data, False)
        customer = user2stripe_customer(expand=['data.subscriptions'])

        if customer.subscriptions.data:
            plan = customer.subscriptions.data[0].plan
            subscription = plan.nickname or plan.id
        else:
            subscription = ''
        session = Session.create(
            api_key=ca.config['STRIPE_SECRET_KEY'],
            **sh.combine_nested_dicts(data, base={
                "customer": customer.id,
                'return_url': request.referrer,
                'locale': session.get('locale', 'en_US').split('_')[0]
            })
        )
    except Exception as e:
        return jsonify(error=str(e))

    return jsonify(session_url=session.url, subscription=subscription)


def get_discounts():
    discounts = {}
    for k, v in sh.stack_nested_keys(get_wallet(cu.id).subscription()):
        if k[-1] == 'discounts':
            for product, flat, perc in json.loads(v):
                f, p = discounts.get(product, (0, 1))
                discounts[product] = f + flat, p * (1 - perc)
    discounts = {k: list(v) for k, v in discounts.items() if v != (0, 1)}
    if discounts:
        from flask import current_app as ca
        api_key = ca.config['STRIPE_SECRET_KEY']
        price_discounts = {}
        product_discounts = {}
        for prod, name in (
                (product, product.name)
                for product in stripe.Product.list(
            active=True, api_key=api_key
        ).auto_paging_iter() if product.name in discounts):
            product_discounts[prod.id] = name
            for price in stripe.Price.list(
                    active=True, product=prod.id, api_key=api_key
            ).auto_paging_iter():
                price_discounts[price.id] = name
        return {
            'discounts': discounts,
            'prod_name': {k: k for k in discounts},
            'price': price_discounts,
            'product': product_discounts
        }
    return {}


def update_line_items_discounts(line_items, discounts):
    from flask import current_app as ca
    api_key = ca.config['STRIPE_SECRET_KEY']
    line_items = copy.deepcopy(line_items)
    for item in line_items:
        if 'price' in item and item['price'] in discounts['price']:
            p = stripe.Price.retrieve(item.pop('price'), api_key=api_key)
            item['price_data'] = {
                'currency': p.currency,
                'product': p.product.id,
                'recurring': p.recurring,
                'tax_behavior': p.tax_behavior,
                'unit_amount_decimal': p.unit_amount_decimal
            }
        if 'price_data' not in item:
            continue
        price_data = item['price_data']
        if 'product' in price_data:
            d = discounts['product'].get(price_data['product'])
        else:
            d = discounts['prod_name'].get(price_data['product_data']['name'])
        if d is None:
            continue
        d = discounts['discounts'][d]
        for k, s in (('unit_amount', 1.0), ('unit_amount_decimal', 100.0)):
            if k not in price_data:
                continue
            if d[0]:
                quantity = item['quantity']
                cost = float(price_data[k]) / s * quantity
                new_cost = max(cost - d[0], 0)
                d[0] -= cost - new_cost
                amount = new_cost / quantity * s
            else:
                amount = float(price_data[k])

            price_data[k] = '%d' % math.ceil(amount * d[1])
    return line_items


def format_line_items(line_items):
    discounts = get_discounts()
    line_items = copy.deepcopy(line_items)
    lookup_keys = {}
    for i, d in enumerate(line_items):
        lookup_key = d.pop('lookup_key', None)
        if lookup_key:
            sh.get_nested_dicts(
                lookup_keys, lookup_key, default=list
            ).append(i)
    if lookup_keys:
        from flask import current_app as ca
        api_key = ca.config['STRIPE_SECRET_KEY']
        for price in stripe.Price.list(
                active=True, api_key=api_key, expand=['data.product'],
                lookup_keys=list(lookup_keys.keys())
        ).auto_paging_iter():
            discount = discounts.get('prod_name', {}).get(price.product.name)
            for i in lookup_keys[price.lookup_key]:
                item = line_items[i]
                if discount is None:
                    item['price'] = price.id
                else:
                    item['price_data'] = {
                        'currency': price.currency,
                        'product': price.product.id,
                        'recurring': price.recurring,
                        'tax_behavior': price.tax_behavior,
                        'unit_amount_decimal': price.unit_amount_decimal,
                    }
    if discounts:
        return update_line_items_discounts(line_items, discounts)
    return line_items


@bp.route('/create-checkout-session', methods=['POST'])
def create_payment():
    from stripe.checkout import Session
    from flask import request, current_app as ca, session
    try:
        data = request.get_json() if request.is_json else dict(request.form)
        data = json_secrets.secrets(data, False)
        if data['mode'] == 'subscription':
            customer = user2stripe_customer(expand=['data.subscriptions'])
            if any(v.status == 'active' for v in customer.subscriptions.data):
                return create_portal(True)
        else:
            customer = user2stripe_customer()

        metadata = {f'customer_{k}': getattr(cu, k) for k in (
            'id', 'firstname', 'lastname', 'email'
        ) if hasattr(cu, k)}
        metadata.update({f'customer_{k}': json.dumps(getattr(cu, k)) for k in (
            'custom_data',
        ) if hasattr(cu, k)})

        api_key = ca.config['STRIPE_SECRET_KEY']

        if 'line_items' in data:
            it = data['line_items']
            if not isinstance(it, list):
                it = [it]
            line_items = []
            for d in it:
                if 'tiers' in d:
                    line_items.extend(compute_line_items(
                        d['quantity'], **d.pop('tiers')
                    ))
                else:
                    line_items.append(d)
            line_items = format_line_items(line_items)
            metadata['line_items'] = json.dumps([
                d.pop('metadata', None) for d in line_items
            ])
            data['line_items'] = line_items

        session = Session.create(
            api_key=api_key,
            **sh.combine_nested_dicts(data, base={
                'ui_mode': 'embedded',
                'customer': customer.id,
                'customer_update': {"address": "auto"},
                'automatic_tax': {'enabled': True},
                'redirect_on_completion': 'never',
                'metadata': metadata,
                'locale': session.get('locale', 'en_US').split('_')[0]
            })
        )
    except Exception as e:
        return jsonify(error=str(e))

    return jsonify(clientSecret=session.client_secret, sessionId=session.id)


def checkout_session_completed(session_id):
    from flask import current_app as ca
    from stripe.checkout import Session
    with Lock(f'Txn-stripe-{session_id}'):
        if db.session.query(
                Txn.query.filter_by(stripe_id=session_id).exists()
        ).scalar():
            return False
        session = Session.retrieve(
            session_id, api_key=ca.config['STRIPE_SECRET_KEY'],
            expand=['line_items.data.price.product', 'customer']
        )
        if session.mode != 'payment':
            return
        customer = session.customer
        current_time = datetime.datetime.fromtimestamp(session.created)
        from asteval import Interpreter
        aeval = Interpreter(usersyms={
            'now': current_time,
            'relativedelta': relativedelta
        }, minimal=True)

        user = stripe_customer2user(customer)
        wallet = get_wallet(user.id)

        line_items = json.loads(session.metadata.get('line_items', '[]'))
        transactions = []
        for i, item in enumerate(session.line_items.data):
            price = item.price
            product = price.product
            expired_at = aeval(price.metadata.get(
                'expires_at', product.metadata.get('expires_at', 'None')
            ))
            try:
                credits = line_items[i]['credits']
            except (IndexError, KeyError):
                credits = item.quantity

            transactions.append(Txn(
                wallet_id=wallet.id,
                type_id=PURCHASE,
                product=product.name,
                subtotal=item.amount_subtotal,
                discount=item.amount_discount,
                tax=item.amount_tax,
                total=item.amount_total,
                currency=item.currency,
                stripe_id=session_id,
                raw_data=item.to_dict_recursive(),
                created_by=user.id,
                valid_from=current_time,
                expired_at=expired_at,
            ))
            transactions.append(Txn(
                wallet_id=wallet.id,
                type_id=CHARGE,
                credits=credits,
                product=product.name,
                stripe_id=session_id,
                created_by=user.id,
                valid_from=current_time,
                expired_at=expired_at,
            ))
        db.session.add_all(transactions)
        db.session.commit()
        return True


def refund_charge(stripe_id, start_time, session, type_ids=(CHARGE,)):
    with Lock(f'Txn-stripe-{stripe_id}'):
        base = Txn.query.filter_by(stripe_id=stripe_id).filter(or_(*(
            Txn.type_id == type_id for type_id in type_ids
        )))

        base.filter(Txn.valid_from > start_time).delete(
            synchronize_session=False
        )
        base.filter(or_(
            Txn.expired_at == None, Txn.expired_at > start_time
        )).update({"expired_at": start_time})
        session.commit()


def subscription_invoice_paid(event):
    invoice = event.data.object
    billing_reason = invoice.billing_reason
    if billing_reason not in (
            'subscription_create', 'subscription_update', 'subscription_cycle'
    ):
        return
    from flask import current_app as ca
    with Lock(f'Txn-stripe-{invoice.id}'):
        if db.session.query(
                Txn.query.filter_by(stripe_id=invoice.id).exists()
        ).scalar():
            return False

        api_key = ca.config['STRIPE_SECRET_KEY']
        subscription = stripe.Subscription.retrieve(
            invoice.subscription, api_key=api_key, expand=[
                'customer', 'items.data.price.product'
            ]
        )

        customer = subscription.customer
        user = stripe_customer2user(customer)
        wallet = get_wallet(user.id)

        start_time = datetime.datetime.fromtimestamp(
            subscription.current_period_start
        )
        end_time = datetime.datetime.fromtimestamp(
            subscription.current_period_end
        ) + relativedelta(days=1)

        if billing_reason == 'subscription_update':
            latest_invoice = Txn.query.filter_by(
                wallet_id=wallet.id, type_id=SUBSCRIPTION
            ).filter(Txn.valid_from <= start_time).order_by(
                desc(Txn.valid_from)
            ).first()
            if latest_invoice:
                latest_invoice = latest_invoice.stripe_id
                refund_charge(latest_invoice, start_time, db.session, (
                    CHARGE, SUBSCRIPTION
                ))
        transactions = []
        for item in subscription.get('items').data:
            product = item.price.product
            if item.object == 'subscription_item':
                transactions.append(Txn(
                    wallet_id=wallet.id,
                    type_id=SUBSCRIPTION,
                    product=product.name,
                    subtotal=invoice.subtotal,
                    discount=sum((
                        v['amount']
                        for v in invoice.total_discount_amounts or []
                    ), 0),
                    tax=invoice.tax,
                    total=invoice.total,
                    currency=invoice.currency,
                    stripe_id=invoice.id,
                    raw_data=invoice.to_dict_recursive(),
                    created_by=user.id,
                    valid_from=start_time,
                    expired_at=end_time
                ))
                products = json.loads(product.metadata.get('products', '[]'))
                products.extend(
                    json.loads(item.price.metadata.get('products', '[]'))
                )
                for feat in stripe.Product.list_features(
                        product.id, api_key=api_key
                ).data:
                    metadata = feat.entitlement_feature.metadata or {}
                    products.extend(json.loads(metadata.get('products', '[]')))
                for name, credits, freq in products:
                    for valid_from, expired_at in date_range(
                            start_time, end_time, freq
                    ):
                        transactions.append(Txn(
                            wallet_id=wallet.id,
                            type_id=CHARGE,
                            credits=credits,
                            product=name,
                            stripe_id=invoice.id,
                            created_by=user.id,
                            valid_from=valid_from,
                            expired_at=expired_at
                        ))
        db.session.add_all(transactions)
        db.session.commit()
        return True


def charge_refunded(event):
    from sqlalchemy.exc import NoResultFound
    from flask import current_app as ca
    from stripe.checkout import Session
    api_key = ca.config['STRIPE_SECRET_KEY']
    charge = event.data.object
    amount_refunded = charge.amount_refunded

    try:
        stripe_id = charge.invoice
        wallet_id = Txn.query.filter_by(
            stripe_id=stripe_id, type_id=SUBSCRIPTION
        ).one().wallet_id
    except NoResultFound:
        try:
            stripe_id = Session.list(
                payment_intent=charge.payment_intent, api_key=api_key, limit=1
            ).data[0].id
            wallet_id = Txn.query.filter_by(
                stripe_id=stripe_id, type_id=PURCHASE
            ).first().wallet_id
        except (IndexError, AttributeError):
            return

    current_time = datetime.datetime.fromtimestamp(event.created)

    refund_charge(stripe_id, current_time, db.session)
    if amount_refunded:
        db.session.add(Txn(
            type_id=REFUND,
            stripe_id=event.id,
            wallet_id=wallet_id,
            total=charge.amount_refunded,
            currency=charge.currency,
            raw_data=charge.to_dict_recursive(),
            valid_from=current_time,
        ))
        db.session.commit()


@bp.route('/session-status/<session_id>', methods=['GET'])
def session_status(session_id):
    from flask import current_app as ca
    from stripe.checkout import Session
    session = Session.retrieve(
        session_id, api_key=ca.config['STRIPE_SECRET_KEY']
    )
    status = session.status
    if status == "complete":
        msg = lazy_gettext('Payment succeeded!', domain='credits')
        category = 'success'
        checkout_session_completed(session_id)
    elif status == "processing":
        msg = lazy_gettext('Your payment is processing.', domain='credits')
        category = 'info'
    elif status == "requires_payment_method":
        msg = lazy_gettext(
            'Your payment was not successful, please try again.',
            domain='credits'
        )
        category = 'warning'
    else:
        msg = lazy_gettext('Something went wrong.', domain='credits')
        category = 'error'
    flash(str(msg), category)
    return jsonify(
        status=status,
        customer_email=session.customer_details.email,
        userInfo=getattr(cu, "get_security_payload", lambda: {})()
    )


@bp.route('/webhooks', methods=['POST'])
@csrf.exempt
def stripe_webhook():
    from flask import request, current_app as ca
    payload = request.data
    sig_header = request.headers['STRIPE_SIGNATURE']
    api_key = ca.config['STRIPE_SECRET_KEY']

    try:
        event = stripe.Webhook.construct_event(
            payload, sig_header,
            ca.config['STRIPE_WEBHOOK_SECRET_KEY'],
            api_key=api_key,
            tolerance=None
        )
    except ValueError as e:
        # Invalid payload
        raise e
    except stripe.error.SignatureVerificationError as e:
        # Invalid signature
        raise e
    event_type = event.type
    if event_type == 'checkout.session.completed':
        checkout_session_completed(event.data.object.id)
    elif event_type == 'charge.refunded':
        charge_refunded(event)
    elif event_type == 'invoice.paid':
        subscription_invoice_paid(event)

    ca.stripe_event_handler(event)
    return jsonify(success=True)


class Credits:
    def __init__(self, app, sitemap, *args, **kwargs):
        if app is not None:
            self.init_app(app, sitemap, *args, **kwargs)

    def init_app(self, app, sitemap, *args, **kwargs):
        app.extensions = getattr(app, 'extensions', {})
        for k in (
                'STRIPE_SECRET_KEY', 'STRIPE_PUBLISHABLE_KEY',
                'STRIPE_WEBHOOK_SECRET_KEY'
        ):
            app.config[k] = app.config.get(k, os.environ.get(k))
            assert app.config[k], f'`{k}` is required!'
        app.stripe_event_handler = sitemap.stripe_event_handler
        app.register_blueprint(bp, url_prefix='/stripe')
        app.extensions['schedula_credits'] = self
        import sherlock
        lock_config = sherlock._configuration
        try:
            lock_config.client
        except ValueError as ex:
            if lock_config.backend is None:
                sherlock.configure(backend=sherlock.backends.FILE)
            else:
                raise ex
        if 'schedula_admin' in app.extensions:
            admin = app.extensions['schedula_admin']
            for v in (Wallet, Txn, TxnType):
                admin.add_model(v, category="Credits")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2025, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
It provides dispatcher sphinx documenter and directive.

Extensions:

.. currentmodule:: schedula.ext.dispatcher

.. autosummary::
    :nosignatures:
    :toctree: dispatcher/

    documenter
    graphviz
"""
from .graphviz import setup as setup_graphviz
from .documenter import setup as setup_documenter, PLOT


def setup(app):
    """Setup `dispatcher` Sphinx extension module. """
    setup_graphviz(app)
    setup_documenter(app)
